import math
import typing
from collections import Counter

import lightgbm as lgb
import numpy as np
import optuna
import pandas as pd
import scipy.stats as ss
from boostaroota import BoostARoota
from lightgbm import LGBMClassifier
from numpy import dot
from numpy.linalg import norm
from sklearn.decomposition import PCA
from sklearn.linear_model import RidgeClassifier
from sklearn.metrics import matthews_corrcoef
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import SVC


class DEESCClassifier:
    """
    Base implementation for delta embedding extended stacking classifier.
    """

    def __init__(
        self,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        Y_train,
        Y_test,
        learning_rate=0.3,
        random_state=1000,
        use_long_warmup: bool = False,
        auto_select_features: bool = False,
        max_tuning_time_h: float = 2,
    ):
        if isinstance(Y_train, pd.Series):
            pass
        else:
            Y_train = pd.Series(Y_train)
        (
            self.X_train,
            self.X_train_2nd_layer,
            self.Y_train,
            self.Y_train_2nd_layer,
        ) = train_test_split(
            X_train, Y_train, test_size=0.70, random_state=random_state
        )
        self.X_test = X_test
        self.Y_test = Y_test
        self.original_columns = [str(col) for col in self.X_train.columns.to_list()]
        self.numerical_columns = self.X_train.select_dtypes(
            include=np.number
        ).columns.tolist()
        self.selected_feats = []  # type: typing.List[str]
        self.selected_feats_2nd_layer = []  # type: typing.List[str]
        self.unique_classes = self.Y_train.unique()
        self.delta_cols_mapping = {}  # type: typing.Dict[str, float]
        self.corr_cols_created = {}  # type: typing.Dict[float, list]
        self.cosine_cols_created = {}  # type: typing.Dict[float, list]
        self.class_corr_cols_created = []  # type: typing.List[str]
        self.class_cosine_cols_created = []  # type: typing.List[str]
        self.cosine_features_selected = {}  # type: typing.Dict[float, list]
        self.correlation_features_selected = {}  # type: typing.Dict[float, list]
        for ind_class in self.unique_classes:
            self.corr_cols_created[ind_class] = []
            self.cosine_cols_created[ind_class] = []
        self.meta_data = {}  # type: typing.Dict[float, pd.DataFrame]
        self.learning_rate = learning_rate
        self.weights_array = self.learning_rate * np.arange(-100, 100)
        self.n_jobs = 1
        self.max_tuning_time_h = max_tuning_time_h
        self.use_long_warmup = use_long_warmup
        self.auto_select_features = auto_select_features
        self.stats_to_consider = ["min", "25%", "50%", "75%", "mean", "max"]
        self.ext_stats_to_consider = [
            "min",
            "1%",
            "5%",
            "10%",
            "20%",
            "25%",
            "50%",
            "75%",
            "80%",
            "90%",
            "95%",
            "99%",
            "mean",
            "max",
        ]
        self.df_shapes = {
            "train_N": self.X_train.shape[0],
            "train_D": self.X_train.shape[1],
            "test_N": self.X_test.shape[0],
            "test_D": self.X_test.shape[1],
        }
        self.delta_cols_created = []  # type: typing.List[str]
        self.base_weights = {}  # type: typing.Dict[float, np.ndarray]
        for ind_class in self.unique_classes:
            self.base_weights[ind_class] = {}
            for stat in self.stats_to_consider:
                self.base_weights[ind_class][stat] = np.ones(self.df_shapes["train_D"])

        self.best_weights = {}  # type: typing.Dict[float, np.ndarray]
        for ind_class in self.unique_classes:
            self.best_weights[ind_class] = {}
            for stat in self.stats_to_consider:
                self.best_weights[ind_class][stat] = np.ones(self.df_shapes["train_D"])

        self.best_score = 0
        self.best_train_score = 0
        self.random_state = random_state
        self.lgbm_model = LGBMClassifier()
        self.ridge_model = RidgeClassifier()
        self.svm_model = SVC()
        self.lgbm_2nd_meta_model = None
        self.lgbm_cols_created = []  # type: typing.List[str]
        self.ridge_cols_created = []  # type: typing.List[str]
        self.svm_cols_created = []  # type: typing.List[str]
        self.lgbm_predictions_non_delta = []  # type: typing.List[str]
        self.ridge_predictions_non_delta = []  # type: typing.List[str]
        self.svm_predictions_non_delta = []  # type: typing.List[str]
        self.nearest_class_predictions = None
        self.class_correlation_predictions = None
        self.class_cosine_predictions = None
        self.lgbm_predictions = None
        self.ridge_predictions = None
        self.svm_predictions = None
        self.lgbm_final_predictions = []  # type: ignore
        self.scaler = MinMaxScaler()
        self.pca_encoder = PCA()
        self.pca_components = []  # type: typing.List[str]

    def add_weights(
        self, delta_matrix, weight: float, col_idx: int, lookup_class, stat
    ):
        """
        Takes a numpy array and adds weights to the specified column_index.
        :param weight: Float. Column value will be multiplied by this value.
        :param col_idx: Integer providing numeric column index.
        :return: Returns modified numpy array.
        """
        temp_weights = self.base_weights[lookup_class][stat]
        temp_weights = np.nan_to_num(temp_weights, neginf=0)
        temp_weights[col_idx : col_idx + 1] = (
            temp_weights[col_idx : col_idx + 1] * weight
        )
        delta_matrix = delta_matrix * temp_weights
        return delta_matrix, temp_weights

    def get_meta_data(self):
        """
        Loops through all classes of a DataFrame, slices it and stores the metadata in a class self.meta_data attribute.
        :return: Updates class attribute.
        """
        for ind_class in self.unique_classes:
            X_train_class_only = self.X_train[self.original_columns].iloc[
                np.where(self.Y_train == ind_class)[0]
            ]
            meta_data = X_train_class_only.describe(
                percentiles=[
                    0.01,
                    0.05,
                    0.10,
                    0.20,
                    0.25,
                    0.5,
                    0.75,
                    0.80,
                    0.90,
                    0.95,
                    0.99,
                ]
            )
            self.meta_data[ind_class] = meta_data

    def delta_to_stat(self, dataframe: pd.DataFrame, stat: str, lookup_class):
        """
        Takes the desired stat as string and the class to look up for. Takes the DataFrame and calculates the delta
        between each row and the class stat.
        :param stat: Str. Must be present in pd.DataFrame.describe() (i.e. "mean", "min", "max")
        :param lookup_class: Class that has been used as key in self.meta_data attribute
        :return: Returns numpy array with all deltas of shape equal to original DataFrame
        """
        delta_to_stat = (
            dataframe.iloc[:, :].values
            - self.meta_data[lookup_class].loc[stat, :].values
        )
        return delta_to_stat

    def get_matthews(self, y_original, preds):
        matthews = matthews_corrcoef(y_original, preds)
        return matthews

    def corr2_coeff_rowwise(self, A, B):
        # Rowwise mean of input arrays & subtract from input arrays themeselves
        A_mA = A - A.mean(1)[:, None]
        B_mB = B - B.mean(1)[:, None]
        # Sum of squares across rows
        ssA = (A_mA ** 2).sum(1)
        ssB = (B_mB ** 2).sum(1)
        # Finally get corr coeff
        return np.einsum("ij,ij->i", A_mA, B_mB) / np.sqrt(ssA * ssB)

    def cosine_similarity(self, a, b):
        return dot(a, b) / (norm(a) * norm(b))

    def conditional_entropy(self, x, y):
        # entropy of x given y
        y_counter = Counter(y)
        xy_counter = Counter(list(zip(x, y)))
        total_occurrences = sum(y_counter.values())
        entropy = 0
        for xy in xy_counter.keys():
            p_xy = xy_counter[xy] / total_occurrences
            p_y = y_counter[xy[1]] / total_occurrences
            entropy += p_xy * math.log(p_y / p_xy)
        return entropy

    def theil_u(self, x, y):
        s_xy = self.conditional_entropy(x, y)
        x_counter = Counter(x)
        total_occurrences = sum(x_counter.values())
        p_x = list(map(lambda n: n / total_occurrences, x_counter.values()))
        s_x = ss.entropy(p_x)
        if s_x == 0:
            return 1
        else:
            return (s_x - s_xy) / s_x

    def theil_u_based_feature_selection(self):
        # transform numerical columns into bins to categorize them
        X_train_temp = self.X_train.copy()
        for col in self.X_train[self.numerical_columns].columns.to_list():
            X_train_temp[str(col)] = pd.qcut(
                X_train_temp[col].rank(method="first"), 10, labels=False
            )
        # get Theil U values
        X_train_temp["target"] = self.Y_train
        theilu = pd.DataFrame(index=["target"], columns=X_train_temp.columns)
        columns = X_train_temp.columns.to_list()
        for j in range(0, len(columns)):
            u = self.theil_u(
                X_train_temp["target"].tolist(), X_train_temp[columns[j]].tolist()
            )
            theilu.loc[:, columns[j]] = u
        theilu.fillna(value=np.nan, inplace=True)

        # feature selection based on Theil U
        for col in theilu.columns.to_list():
            if theilu.loc[:, col].values[0] >= 0.005 and col != "target":
                self.selected_feats.append(col)

    def feature_selection(self):
        br = BoostARoota(metric="mlogloss")
        br.fit(self.X_train, self.Y_train)
        selected = br.keep_vars_.values.tolist()
        selected = [str(col) for col in selected]
        self.selected_feats = selected

    def last_layer_feature_selection(self):
        br = BoostARoota(metric="mlogloss")
        br.fit(self.X_train_2nd_layer, self.Y_train_2nd_layer)
        selected = br.keep_vars_
        selected = selected.values.tolist()
        selected = [str(col) for col in selected]
        self.selected_feats_2nd_layer = selected

    def fit_scale_data(self, dataframe: pd.DataFrame):
        standard_scaler = MinMaxScaler()
        columns = dataframe.columns
        dataframe = pd.DataFrame(
            standard_scaler.fit_transform(dataframe), columns=columns
        )
        self.scaler = standard_scaler
        return dataframe

    def predict_scale_data(self, dataframe: pd.DataFrame):
        standard_scaler = self.scaler
        columns = dataframe.columns
        dataframe = pd.DataFrame(standard_scaler.transform(dataframe), columns=columns)
        return dataframe

    def fit_delta_embeddings_pca(self, dataframe: pd.DataFrame):
        dataframe = self.fit_scale_data(dataframe)
        pca = PCA(
            n_components=0.95,
            random_state=self.random_state,
        )
        train_comps = pca.fit_transform(dataframe)
        new_cols = [f"PCA_{i}" for i in range(train_comps.shape[1])]

        self.pca_encoder = pca
        self.pca_components = new_cols

    def predict_delta_embeddings_pca(self, dataframe: pd.DataFrame):
        dataframe = self.predict_scale_data(dataframe)
        pca = self.pca_encoder
        new_cols = self.pca_components
        components = pca.transform(dataframe)
        reduced_df = pd.DataFrame(components, columns=new_cols)
        return reduced_df

    def create_baseline_score(self):
        self.get_meta_data()
        # initial filling & base line score
        for lookup_class in self.unique_classes:
            all_non_stat_delta_train = np.ones(
                (self.df_shapes["train_N"], self.df_shapes["train_D"]), dtype="float64"
            )
            all_non_stat_delta_test = np.ones(
                (self.df_shapes["test_N"], self.df_shapes["test_D"]), dtype="float64"
            )
            for stat in self.stats_to_consider:
                stat_delta = self.delta_to_stat(
                    self.X_train[self.original_columns], stat, lookup_class
                )
                stat_delta = np.nan_to_num(stat_delta, neginf=0)
                all_non_stat_delta_train += stat_delta
                total_delta_matrix = stat_delta + all_non_stat_delta_train
                total_delta_df = pd.DataFrame(total_delta_matrix)
                self.X_train[
                    f"total_delta_sum_to_class_{lookup_class}"
                ] = total_delta_df.sum(axis=1).values

                stat_delta = self.delta_to_stat(
                    self.X_test[self.original_columns], stat, lookup_class
                )
                all_non_stat_delta_test += stat_delta
                total_delta_matrix = stat_delta + all_non_stat_delta_test
                total_delta_df = pd.DataFrame(total_delta_matrix)
                self.X_test[
                    f"total_delta_sum_to_class_{lookup_class}"
                ] = total_delta_df.sum(axis=1).values
                self.delta_cols_created.append(
                    f"total_delta_sum_to_class_{lookup_class}"
                )
                self.delta_cols_mapping[
                    f"total_delta_sum_to_class_{lookup_class}"
                ] = lookup_class

        self.X_train["supervised_distance_nearest_class"] = (
            self.X_train[self.delta_cols_created].idxmin(axis=1).values
        )
        self.X_test["supervised_distance_nearest_class"] = (
            self.X_test[self.delta_cols_created].idxmin(axis=1).values
        )

        self.X_train["supervised_distance_nearest_class"] = self.X_train[
            "supervised_distance_nearest_class"
        ].map(self.delta_cols_mapping)

        self.X_test["supervised_distance_nearest_class"] = self.X_test[
            "supervised_distance_nearest_class"
        ].map(self.delta_cols_mapping)

        self.best_score = self.get_matthews(
            self.Y_test, self.X_test["supervised_distance_nearest_class"]
        )
        print(f"Baseline score with default weights is {self.best_score}.")

    def warmup_weights(self):
        # optimization loop
        trials = 0
        total_trials = (
            len(self.unique_classes)
            * len(self.weights_array)
            * self.df_shapes["train_D"]
        )
        print(f"Start warmup. Total trials will be: {total_trials}:")
        for lookup_class in self.unique_classes:
            print(f"Start loop for class {lookup_class}...")
            for stat in ["min"]:
                print(f"Start loop for stat {stat}...")
                # print(f"Start loop for stat {stat}")
                non_sel = self.stats_to_consider.copy()
                non_sel.remove(stat)
                for col in range(self.df_shapes["train_D"]):
                    # print(f"Start loop for col {col}")
                    for weight in self.weights_array:
                        stat_delta = self.delta_to_stat(
                            self.X_train[self.original_columns], stat, lookup_class
                        )
                        stat_delta, temp_weights = self.add_weights(
                            stat_delta, weight, col, lookup_class, stat
                        )
                        stat_delta = np.nan_to_num(stat_delta, neginf=0)

                        all_non_stat_delta = np.ones(
                            (self.df_shapes["train_N"], self.df_shapes["train_D"]),
                            dtype="float64",
                        )
                        for non_sel_stat in non_sel:
                            non_sel_stat_delta = self.delta_to_stat(
                                self.X_train[self.original_columns],
                                non_sel_stat,
                                lookup_class,
                            )
                            # non_sel_stat_delta, temp_weights_non = self.add_weights(non_sel_stat_delta, 1, col, lookup_class, non_sel_stat)
                            non_sel_stat_delta = np.nan_to_num(
                                non_sel_stat_delta, neginf=0
                            )
                            all_non_stat_delta += non_sel_stat_delta
                        total_delta_matrix = stat_delta + all_non_stat_delta
                        total_delta_df = pd.DataFrame(total_delta_matrix)
                        self.X_train[
                            f"total_delta_sum_to_class_{lookup_class}"
                        ] = total_delta_df.sum(axis=1).values

                        stat_delta = self.delta_to_stat(
                            self.X_test[self.original_columns], stat, lookup_class
                        )
                        stat_delta, temp_weights_test = self.add_weights(
                            stat_delta, weight, col, lookup_class, stat
                        )
                        stat_delta = np.nan_to_num(stat_delta, neginf=0)
                        all_non_stat_delta = np.ones(
                            (self.df_shapes["test_N"], self.df_shapes["test_D"]),
                            dtype="float64",
                        )
                        all_non_stat_delta = np.nan_to_num(all_non_stat_delta, neginf=0)
                        for non_sel_stat in non_sel:
                            non_sel_stat_delta = self.delta_to_stat(
                                self.X_test[self.original_columns],
                                non_sel_stat,
                                lookup_class,
                            )
                            # non_sel_stat_delta, temp_weights_non = self.add_weights(non_sel_stat_delta, 1, col, lookup_class, non_sel_stat)
                            non_sel_stat_delta = np.nan_to_num(
                                non_sel_stat_delta, neginf=0
                            )
                            all_non_stat_delta += non_sel_stat_delta

                        total_delta_matrix = stat_delta + all_non_stat_delta
                        total_delta_df = pd.DataFrame(total_delta_matrix)
                        self.X_test[
                            f"total_delta_sum_to_class_{lookup_class}"
                        ] = total_delta_df.sum(axis=1).values

                        self.X_train["supervised_distance_nearest_class"] = (
                            self.X_train[self.delta_cols_created].idxmin(axis=1).values
                        )
                        self.X_test["supervised_distance_nearest_class"] = (
                            self.X_test[self.delta_cols_created].idxmin(axis=1).values
                        )

                        self.X_train[
                            "supervised_distance_nearest_class"
                        ] = self.X_train["supervised_distance_nearest_class"].map(
                            self.delta_cols_mapping
                        )

                        self.X_test["supervised_distance_nearest_class"] = self.X_test[
                            "supervised_distance_nearest_class"
                        ].map(self.delta_cols_mapping)

                        trials += 1
                        progress = round((trials / total_trials), 2)

                        matthews_train = self.get_matthews(
                            self.Y_train,
                            self.X_train["supervised_distance_nearest_class"],
                        )
                        matthews = self.get_matthews(
                            self.Y_test,
                            self.X_test["supervised_distance_nearest_class"],
                        )

                        if matthews_train > self.best_train_score:
                            print(
                                f"New best train score after trial {trials} is {matthews_train}."
                            )
                            self.best_train_score = matthews_train

                        if matthews > self.best_score:
                            print(
                                f"New best score after trial {trials} is {matthews}. Best training set score is {self.best_train_score}."
                            )
                            self.best_weights[lookup_class][stat] = temp_weights
                            self.best_score = matthews

                        if trials % 1000 == 0:
                            print(
                                f"Reached progress of {progress * 100}% with current best score of {self.best_score}."
                            )

    def fit_delta_weights(self):
        if self.auto_select_features:
            self.feature_selection()
            if len(self.selected_feats) > 0:
                self.X_train = self.X_train[self.selected_feats].copy()
                self.X_test = self.X_test[self.selected_feats].copy()
                self.original_columns = self.selected_feats
                self.df_shapes = {
                    "train_N": self.X_train.shape[0],
                    "train_D": self.X_train.shape[1],
                    "test_N": self.X_test.shape[0],
                    "test_D": self.X_test.shape[1],
                }
                self.base_weights = {}  # type: typing.Dict[float, np.ndarray]
                for ind_class in self.unique_classes:
                    self.base_weights[ind_class] = {}
                    for stat in self.stats_to_consider:
                        self.base_weights[ind_class][stat] = np.ones(
                            self.df_shapes["train_D"]
                        )

                self.best_weights = {}  # type: typing.Dict[float, np.ndarray]
                for ind_class in self.unique_classes:
                    self.best_weights[ind_class] = {}
                    for stat in self.stats_to_consider:
                        self.best_weights[ind_class][stat] = np.ones(
                            self.df_shapes["train_D"]
                        )
            else:
                pass

        self.create_baseline_score()
        if self.use_long_warmup:
            self.warmup_weights()

        algorithm = "delta_classifier"
        max_trials = (
            len(self.unique_classes)
            * len(self.stats_to_consider)
            * self.df_shapes["train_D"]
            * 100
        )
        nb_trials = int(max_trials / 10)
        warm_up_trials = int(max_trials / 100)
        if warm_up_trials < 1:
            warm_up_trials = 10
        else:
            pass
        print(
            f"The maximum number of trials planned is {nb_trials} with {warm_up_trials} random warmup trials at start."
        )

        def objective(trial):
            lookup_class = trial.suggest_categorical(
                "lookup_class", [cls for cls in self.unique_classes]
            )
            weight = trial.suggest_uniform("weight", -100, 100)
            col = trial.suggest_int(
                "col", 0, len(self.X_train[self.original_columns].columns)
            )
            stat = trial.suggest_categorical("stat", self.stats_to_consider)
            non_sel = self.stats_to_consider.copy()
            non_sel.remove(str(stat))
            stat_delta = self.delta_to_stat(
                self.X_train[self.original_columns], stat, lookup_class
            )
            stat_delta, temp_weights = self.add_weights(
                stat_delta, weight, col, lookup_class, stat
            )
            stat_delta = np.nan_to_num(stat_delta, neginf=0)
            all_non_stat_delta = np.ones(
                (self.df_shapes["train_N"], self.df_shapes["train_D"]), dtype="float64"
            )
            for non_sel_stat in non_sel:
                non_sel_stat_delta = self.delta_to_stat(
                    self.X_train[self.original_columns], non_sel_stat, lookup_class
                )
                # non_sel_stat_delta, temp_weights_non = self.add_weights(non_sel_stat_delta, 1, col, lookup_class, non_sel_stat)
                non_sel_stat_delta = np.nan_to_num(non_sel_stat_delta, neginf=0)
                all_non_stat_delta += non_sel_stat_delta
            total_delta_matrix = stat_delta + all_non_stat_delta
            total_delta_df = pd.DataFrame(total_delta_matrix)
            self.X_train[
                f"total_delta_sum_to_class_{lookup_class}"
            ] = total_delta_df.sum(axis=1).values
            stat_delta = self.delta_to_stat(
                self.X_test[self.original_columns], stat, lookup_class
            )
            stat_delta, temp_weights_test = self.add_weights(
                stat_delta, weight, col, lookup_class, stat
            )
            stat_delta = np.nan_to_num(stat_delta, neginf=0)
            all_non_stat_delta = np.ones(
                (self.df_shapes["test_N"], self.df_shapes["test_D"]), dtype="float64"
            )
            all_non_stat_delta = np.nan_to_num(all_non_stat_delta, neginf=0)
            for non_sel_stat in non_sel:
                non_sel_stat_delta = self.delta_to_stat(
                    self.X_test[self.original_columns], non_sel_stat, lookup_class
                )
                # non_sel_stat_delta, temp_weights_non = self.add_weights(non_sel_stat_delta, 1, col, lookup_class, non_sel_stat)
                non_sel_stat_delta = np.nan_to_num(non_sel_stat_delta, neginf=0)
                all_non_stat_delta += non_sel_stat_delta
            total_delta_matrix = stat_delta + all_non_stat_delta
            total_delta_df = pd.DataFrame(total_delta_matrix)
            self.X_test[
                f"total_delta_sum_to_class_{lookup_class}"
            ] = total_delta_df.sum(axis=1).values
            self.X_train["supervised_distance_nearest_class"] = (
                self.X_train[self.delta_cols_created].idxmin(axis=1).values
            )
            self.X_test["supervised_distance_nearest_class"] = (
                self.X_test[self.delta_cols_created].idxmin(axis=1).values
            )
            self.X_train["supervised_distance_nearest_class"] = self.X_train[
                "supervised_distance_nearest_class"
            ].map(self.delta_cols_mapping)
            self.X_test["supervised_distance_nearest_class"] = self.X_test[
                "supervised_distance_nearest_class"
            ].map(self.delta_cols_mapping)
            matthews_train = self.get_matthews(
                self.Y_train, self.X_train["supervised_distance_nearest_class"]
            )
            matthews = self.get_matthews(
                self.Y_test, self.X_test["supervised_distance_nearest_class"]
            )
            if matthews_train > self.best_train_score:
                self.best_train_score = matthews_train
            if matthews > self.best_score:
                self.best_weights[lookup_class][stat] = temp_weights
                self.best_score = matthews
            return matthews

        sampler = optuna.samplers.TPESampler(
            multivariate=True, n_startup_trials=20, seed=42
        )
        study = optuna.create_study(
            direction="maximize", study_name=f"{algorithm} tuning", sampler=sampler
        )

        study.optimize(
            objective,
            n_trials=nb_trials,
            timeout=self.max_tuning_time_h * 60 * 60,
            gc_after_trial=True,
            show_progress_bar=True,
            n_jobs=self.n_jobs,
        )

        try:
            fig = optuna.visualization.plot_optimization_history(study)
            fig.show()
        except ZeroDivisionError:
            print(
                "Plotting of hyperparameter performances failed. This usually implicates an error during training."
            )

        try:
            fig = optuna.visualization.plot_param_importances(study)
            fig.show()
        except ZeroDivisionError:
            print(
                "Plotting of hyperparameter performances failed. This usually implicates an error during training."
            )

    def delta_correlations(
        self, dataframe: pd.DataFrame, delta_array, stat, lookup_class
    ):
        df1 = dataframe.rank(axis=1)
        dataframe[f"corr_{lookup_class}_{stat}"] = df1.corrwith(
            pd.Series(delta_array, index=df1.columns).rank(), axis=1
        )
        self.corr_cols_created[lookup_class].append(f"corr_{lookup_class}_{stat}")
        return dataframe[f"corr_{lookup_class}_{stat}"]

    def fit_cosine_similarity(self):
        X_train = self.X_train.copy()
        X_test = self.X_test.copy()

        cosine_cols_wo_class = []
        cosine_cols_to_full_name = {}
        for lookup_class in self.unique_classes:
            for stat in self.ext_stats_to_consider:
                X_train[
                    f"cosine_similarity_{lookup_class}_{stat}"
                ] = self.cosine_similarity(
                    X_train[self.original_columns].values,
                    self.meta_data[lookup_class].loc[stat, :].values,
                )
                X_test[
                    f"cosine_similarity_{lookup_class}_{stat}"
                ] = self.cosine_similarity(
                    X_test[self.original_columns].values,
                    self.meta_data[lookup_class].loc[stat, :].values,
                )
                self.cosine_cols_created[lookup_class].append(
                    f"cosine_similarity_{lookup_class}_{stat}"
                )
                if f"cosine_similarity_{stat}" not in cosine_cols_wo_class:
                    cosine_cols_wo_class.append(f"cosine_similarity_{stat}")
                cosine_cols_to_full_name[
                    f"cosine_similarity_{lookup_class}_{stat}"
                ] = f"cosine_similarity_{stat}"

        def objective(trial):
            param = {}
            # we go crazy here and allow Optuna to chose the columns to consider for clustering
            for col in self.ext_stats_to_consider:
                param[col] = trial.suggest_int(col, 0, 1)

            temp_features_dic = {}
            for lookup_class in self.unique_classes:
                for k, v in param.items():
                    if v == 1:
                        temp_features_dic[
                            lookup_class
                        ] = f"cosine_similarity_{lookup_class}_{k}"
                    else:
                        pass

            temp_mean_cols_mapping = {}
            temp_mean_cols = []
            for lookup_class in self.unique_classes:
                temp_features = [
                    v for k, v in temp_features_dic.items() if k == lookup_class
                ]
                X_train[f"test_score_{lookup_class}"] = X_train[temp_features].mean(
                    axis=1
                )
                X_test[f"test_score_{lookup_class}"] = X_test[temp_features].mean(
                    axis=1
                )
                temp_mean_cols_mapping[f"test_score_{lookup_class}"] = lookup_class
                temp_mean_cols.append(f"test_score_{lookup_class}")

            X_train["final_test_score"] = X_train[temp_mean_cols].idxmax(axis=1).values
            X_train["final_test_score"] = X_train["final_test_score"].map(
                temp_mean_cols_mapping
            )
            X_test["final_test_score"] = X_test[temp_mean_cols].idxmax(axis=1).values
            X_test["final_test_score"] = X_test["final_test_score"].map(
                temp_mean_cols_mapping
            )

            X_train["final_test_score"] = X_train["final_test_score"].fillna(0)
            X_test["final_test_score"] = X_test["final_test_score"].fillna(0)
            matthews_train = matthews_corrcoef(
                self.Y_train, X_train["final_test_score"]
            )
            matthews_test = matthews_corrcoef(self.Y_test, X_test["final_test_score"])
            matthews = matthews_train - (matthews_train - matthews_test) ** 2
            return matthews

        sampler = optuna.samplers.TPESampler(multivariate=True, seed=42)
        study = optuna.create_study(
            direction="maximize", sampler=sampler, study_name="tune_cosine"
        )
        study.optimize(
            objective,
            n_trials=500,
            gc_after_trial=True,
            show_progress_bar=True,
            timeout=1 * 60 * 60,
        )

        final_cosine_features_dic = {}
        for lookup_class in self.unique_classes:
            final_cosine_features_dic[lookup_class] = []
            for k, v in study.best_trial.params.items():
                if v == 1:
                    final_cosine_features_dic[lookup_class].append(
                        f"cosine_similarity_{lookup_class}_{k}"
                    )
                else:
                    pass

        self.cosine_features_selected = final_cosine_features_dic

    def fit_correlations(self):
        X_train = self.X_train.copy()
        X_test = self.X_test.copy()

        corr_cols_wo_class = []
        corr_cols_to_full_name = {}
        for lookup_class in self.unique_classes:
            for stat in self.ext_stats_to_consider:
                X_train[f"corr_{lookup_class}_{stat}"] = self.delta_correlations(
                    X_train[self.original_columns],
                    self.meta_data[lookup_class].loc[stat, :].values,
                    stat,
                    lookup_class,
                )
                X_test[f"corr_{lookup_class}_{stat}"] = self.delta_correlations(
                    X_test[self.original_columns],
                    self.meta_data[lookup_class].loc[stat, :].values,
                    stat,
                    lookup_class,
                )
                self.corr_cols_created[lookup_class].append(
                    f"corr_{lookup_class}_{stat}"
                )
                if f"corr_{lookup_class}_{stat}" not in corr_cols_wo_class:
                    corr_cols_wo_class.append(f"corr_{lookup_class}_{stat}")
                corr_cols_to_full_name[f"corr_{lookup_class}_{stat}"] = f"corr_{stat}"

        def objective(trial):
            param = {}
            # we go crazy here and allow Optuna to chose the columns to consider for clustering
            for col in self.ext_stats_to_consider:
                param[col] = trial.suggest_int(col, 0, 1)

            temp_features_dic = {}
            for lookup_class in self.unique_classes:
                for k, v in param.items():
                    if v == 1:
                        temp_features_dic[lookup_class] = f"corr_{lookup_class}_{k}"
                    else:
                        pass

            temp_mean_cols_mapping = {}
            temp_mean_cols = []
            for lookup_class in self.unique_classes:
                temp_features = [
                    v for k, v in temp_features_dic.items() if k == lookup_class
                ]
                X_train[f"test_score_{lookup_class}"] = X_train[temp_features].mean(
                    axis=1
                )
                X_test[f"test_score_{lookup_class}"] = X_test[temp_features].mean(
                    axis=1
                )
                temp_mean_cols_mapping[f"test_score_{lookup_class}"] = lookup_class
                temp_mean_cols.append(f"test_score_{lookup_class}")

            X_train["final_test_score"] = X_train[temp_mean_cols].idxmax(axis=1).values
            X_train["final_test_score"] = X_train["final_test_score"].map(
                temp_mean_cols_mapping
            )
            X_test["final_test_score"] = X_test[temp_mean_cols].idxmax(axis=1).values
            X_test["final_test_score"] = X_test["final_test_score"].map(
                temp_mean_cols_mapping
            )
            X_train["final_test_score"] = X_train["final_test_score"].fillna(0)
            X_test["final_test_score"] = X_test["final_test_score"].fillna(0)
            matthews_train = matthews_corrcoef(
                self.Y_train, X_train["final_test_score"]
            )
            matthews_test = matthews_corrcoef(self.Y_test, X_test["final_test_score"])
            matthews = matthews_train - (matthews_train - matthews_test) ** 2
            return matthews

        sampler = optuna.samplers.TPESampler(multivariate=True, seed=42)
        study = optuna.create_study(
            direction="maximize", sampler=sampler, study_name="tune_correlations"
        )
        study.optimize(
            objective,
            n_trials=500,
            gc_after_trial=True,
            show_progress_bar=True,
            timeout=1 * 60 * 60,
        )

        final_corr_features_dic = {}
        for lookup_class in self.unique_classes:
            final_corr_features_dic[lookup_class] = []
            for k, v in study.best_trial.params.items():
                if v == 1:
                    final_corr_features_dic[lookup_class].append(
                        f"cosine_similarity_{lookup_class}_{k}"
                    )
                else:
                    pass

        self.correlation_features_selected = final_corr_features_dic

    def fit_base_models(self):
        X_train = self.X_train.copy()

        total_delta_dfs = []
        for lookup_class in self.unique_classes:
            total_delta_matrix = np.ones(
                (
                    X_train[self.original_columns].shape[0],
                    X_train[self.original_columns].shape[1],
                ),
                dtype="float64",
            )
            stats_dfs = []
            for stat in self.stats_to_consider:
                stat_delta = self.delta_to_stat(
                    X_train[self.original_columns], stat, lookup_class
                )

                temp_weights = self.best_weights[lookup_class][stat]
                temp_weights = np.nan_to_num(temp_weights, neginf=0)
                stat_delta = stat_delta * temp_weights
                stat_delta = np.nan_to_num(stat_delta, neginf=0)
                total_delta_matrix += stat_delta
                # get sums of all deltas per class
                total_delta_df = pd.DataFrame(total_delta_matrix)
                stats_dfs.append(total_delta_df)
            for i in range(len(stats_dfs)):
                if i == 0:
                    stats_df = stats_dfs[i]
                else:
                    stats_df = stats_df.merge(
                        stats_dfs[i], left_index=True, right_index=True
                    )
            total_delta_dfs.append(stats_df)

        for i in range(len(total_delta_dfs)):
            if i == 0:
                embeddings_df = total_delta_dfs[i]
            else:
                embeddings_df = embeddings_df.merge(
                    total_delta_dfs[i], left_index=True, right_index=True
                )

        N = embeddings_df.shape[0]
        train_sample = embeddings_df.sample(int(N / 2), random_state=self.random_state)
        # val_sample = embeddings_df[(~embeddings_df.index.isin(train_sample.index))]
        embeddings_df.columns = range(embeddings_df.shape[1])
        self.fit_delta_embeddings_pca(train_sample)
        embeddings_df = self.predict_delta_embeddings_pca(embeddings_df)
        lgbm_model = LGBMClassifier(random_state=self.random_state)
        lgbm_model.fit(embeddings_df, self.Y_train)
        self.lgbm_model = lgbm_model

        ridge_model = RidgeClassifier(random_state=self.random_state)
        ridge_model.fit(embeddings_df, self.Y_train)
        self.ridge_model = ridge_model

        svm_model = SVC(random_state=self.random_state, probability=True)
        svm_model.fit(embeddings_df, self.Y_train)
        self.svm_model = svm_model

    def fit_1st_layer(self):
        self.fit_delta_weights()
        self.fit_cosine_similarity()
        self.fit_correlations()
        self.fit_base_models()

    def predict_cosine_similarity(self, dataframe: pd.DataFrame):
        for lookup_class in self.unique_classes:
            features_to_use = self.cosine_features_selected[lookup_class]
            for stat in self.ext_stats_to_consider:
                dataframe[
                    f"cosine_similarity_{lookup_class}_{stat}"
                ] = self.cosine_similarity(
                    dataframe[self.original_columns].values,
                    self.meta_data[lookup_class].loc[stat, :].values,
                )
                self.cosine_cols_created[lookup_class].append(
                    f"cosine_similarity_{lookup_class}_{stat}"
                )
            dataframe[f"mean_stat_delta_cosine_class_{lookup_class}"] = (
                dataframe[features_to_use].mean(axis=1).values
            )
            self.class_cosine_cols_created.append(
                f"mean_stat_delta_cosine_class_{lookup_class}"
            )

            self.delta_cols_mapping[
                f"mean_stat_delta_cosine_class_{lookup_class}"
            ] = lookup_class

        # get final predictions based on highest cosine to delta
        dataframe["cosine_strongest_class"] = (
            dataframe[self.class_cosine_cols_created].idxmax(axis=1).values
        )
        dataframe["cosine_strongest_class"] = dataframe["cosine_strongest_class"].map(
            self.delta_cols_mapping
        )

        self.class_cosine_predictions = dataframe["cosine_strongest_class"]
        return dataframe

    def predict_correlations(self, dataframe: pd.DataFrame):
        for lookup_class in self.unique_classes:
            features_to_use = self.correlation_features_selected[lookup_class]
            for stat in self.ext_stats_to_consider:
                dataframe[f"corr_{lookup_class}_{stat}"] = self.delta_correlations(
                    dataframe[self.original_columns],
                    self.meta_data[lookup_class].loc[stat, :].values,
                    stat,
                    lookup_class,
                )
                self.corr_cols_created[lookup_class].append(
                    f"corr_{lookup_class}_{stat}"
                )
            dataframe[f"mean_stat_delta_correlation_class_{lookup_class}"] = (
                dataframe[features_to_use].mean(axis=1).values
            )
            self.class_corr_cols_created.append(
                f"mean_stat_delta_correlation_class_{lookup_class}"
            )

            self.delta_cols_mapping[
                f"mean_stat_delta_correlation_class_{lookup_class}"
            ] = lookup_class

        # get final predictions based on highest cosine to delta
        dataframe["correlations_strongest_class"] = (
            dataframe[self.class_corr_cols_created].idxmax(axis=1).values
        )
        dataframe["correlations_strongest_class"] = dataframe[
            "correlations_strongest_class"
        ].map(self.delta_cols_mapping)

        self.class_correlation_predictions = dataframe["correlations_strongest_class"]
        return dataframe

    def get_classes_from_probs(self, probs):
        if len(self.unique_classes) == 2:
            predicted_probs = np.asarray([line[1] for line in probs])
            predicted_classes = predicted_probs > 0.5
        else:
            predicted_probs = probs
            predicted_classes = np.asarray(
                [np.argmax(line) for line in predicted_probs]
            )
        return predicted_classes

    def predict_base_models(self, dataframe: pd.DataFrame):
        total_delta_dfs = []
        for lookup_class in self.unique_classes:
            # predict loop
            total_delta_matrix = np.ones(
                (
                    dataframe[self.original_columns].shape[0],
                    dataframe[self.original_columns].shape[1],
                ),
                dtype="float64",
            )
            stats_dfs = []
            for stat in self.stats_to_consider:
                stat_delta = self.delta_to_stat(
                    dataframe[self.original_columns], stat, lookup_class
                )

                temp_weights = self.best_weights[lookup_class][stat]
                temp_weights = np.nan_to_num(temp_weights, neginf=0)
                stat_delta = stat_delta * temp_weights
                stat_delta = np.nan_to_num(stat_delta, neginf=0)
                total_delta_matrix += stat_delta
                # get sums of all deltas per class
                total_delta_df = pd.DataFrame(total_delta_matrix)
                stats_dfs.append(total_delta_df)
            for i in range(len(stats_dfs)):
                if i == 0:
                    stats_df = stats_dfs[i]
                else:
                    stats_df = stats_df.merge(
                        stats_dfs[i], left_index=True, right_index=True
                    )
            total_delta_dfs.append(stats_df)

        for i in range(len(total_delta_dfs)):
            if i == 0:
                embeddings_df = total_delta_dfs[i]
            else:
                embeddings_df = embeddings_df.merge(
                    total_delta_dfs[i], left_index=True, right_index=True
                )
        embeddings_df.columns = range(embeddings_df.shape[1])
        embeddings_df = self.predict_delta_embeddings_pca(embeddings_df)
        lgbm_model = self.lgbm_model
        temp_preds = lgbm_model.predict_proba(embeddings_df)
        for lookup_class in range(self.unique_classes.shape[0]):
            dataframe[f"lgbm_predictions_proba_{lookup_class}"] = temp_preds[
                :, lookup_class
            ].reshape(-1, 1)
        dataframe["lgbm_predictions"] = self.get_classes_from_probs(temp_preds)
        self.lgbm_predictions = dataframe["lgbm_predictions"]
        self.lgbm_cols_created.append("lgbm_predictions")

        ridge_model = self.ridge_model
        dataframe["ridge_predictions"] = ridge_model.predict(embeddings_df)
        self.ridge_predictions = dataframe["ridge_predictions"]
        self.ridge_cols_created.append("ridge_predictions")

        svm_model = self.svm_model
        temp_preds = svm_model.predict_proba(embeddings_df)
        for lookup_class in range(self.unique_classes.shape[0]):
            dataframe[f"svm_predictions_proba_{lookup_class}"] = temp_preds[
                :, lookup_class
            ].reshape(-1, 1)
        dataframe["svm_predictions"] = self.get_classes_from_probs(temp_preds)
        self.svm_predictions = dataframe["svm_predictions"]
        self.svm_cols_created.append("svm_predictions")
        return dataframe

    def predict_1st_layer(self, dataframe: pd.DataFrame):
        dataframe = dataframe[self.original_columns].copy()

        dataframe = self.predict_cosine_similarity(dataframe)
        dataframe = self.predict_correlations(dataframe)
        dataframe = self.predict_base_models(dataframe)

        for lookup_class in self.unique_classes:
            total_delta_matrix = np.ones(
                (
                    dataframe[self.original_columns].shape[0],
                    dataframe[self.original_columns].shape[1],
                ),
                dtype="float64",
            )
            for stat in self.stats_to_consider:
                stat_delta = self.delta_to_stat(
                    dataframe[self.original_columns], stat, lookup_class
                )

                temp_weights = self.best_weights[lookup_class][stat]
                temp_weights = np.nan_to_num(temp_weights, neginf=0)
                stat_delta = stat_delta * temp_weights
                stat_delta = np.nan_to_num(stat_delta, neginf=0)
                total_delta_matrix += stat_delta
            # get sums of all deltas per class
            total_delta_df = pd.DataFrame(total_delta_matrix)
            dataframe[f"total_delta_sum_to_class_{lookup_class}"] = total_delta_df.sum(
                axis=1
            ).values

        # get final predictions based on delta distances and weights
        dataframe["supervised_distance_nearest_class"] = (
            dataframe[self.delta_cols_created].idxmin(axis=1).values
        )
        dataframe["supervised_distance_nearest_class"] = dataframe[
            "supervised_distance_nearest_class"
        ].map(self.delta_cols_mapping)

        self.nearest_class_predictions = dataframe["supervised_distance_nearest_class"]

        final_predictors = [
            "correlations_strongest_class",
            "cosine_strongest_class",
            "supervised_distance_nearest_class",
            # "lgbm_predictions",
            # "ridge_predictions",
            # "svm_predictions"
        ]

        dataframe = dataframe.fillna(0)
        dataframe["delta_based_predictions"] = dataframe[final_predictors].mode(axis=1)[
            0
        ]
        return dataframe

    def fit(self):
        self.fit_1st_layer()
        self.last_layer_feature_selection()
        dataframe = self.predict_1st_layer(self.X_train_2nd_layer)

        X_train = dataframe[self.selected_feats_2nd_layer].copy()
        Y_train = self.Y_train_2nd_layer.copy()
        X_train["target"] = self.Y_train_2nd_layer

        length = len(X_train.index)

        x_train = X_train.sample(
            int(length / 10 * 8), random_state=self.random_state
        ).copy()
        y_train = x_train["target"]

        X_test = X_train[(~X_train.index.isin(x_train.index))].copy()
        Y_test = X_test["target"]

        x_train = x_train.drop("target", axis=1)
        X_test = X_test.drop("target", axis=1)
        X_train = X_train.drop("target", axis=1)

        dtrain = lgb.Dataset(x_train, label=y_train)

        if len(self.unique_classes) == 2:
            # weights_for_lgb = self.calc_scale_pos_weight()

            def objective(trial):
                param = {
                    "objective": "binary",
                    "metric": "binary_logloss",
                    # 'scale_pos_weight': trial.suggest_loguniform('scale_pos_weight', 1e-3, 1e3),
                    "num_boost_round": trial.suggest_int("num_boost_round", 100, 50000),
                    "lambda_l1": trial.suggest_loguniform("lambda_l1", 1, 1e6),
                    "lambda_l2": trial.suggest_loguniform("lambda_l2", 1, 1e6),
                    # 'max_depth': trial.suggest_int('max_depth', 2, 8),
                    "num_leaves": trial.suggest_int("num_leaves", 2, 256),
                    "feature_fraction": trial.suggest_uniform(
                        "feature_fraction", 0.4, 1.0
                    ),
                    "feature_fraction_bynode": trial.suggest_uniform(
                        "feature_fraction_bynode", 0.4, 1.0
                    ),
                    # 'bagging_freq': trial.suggest_int('bagging_freq', 1, 7),
                    # 'min_child_samples': trial.suggest_int('min_child_samples', 5, 100),
                    "learning_rate": trial.suggest_loguniform(
                        "learning_rate", 1e-5, 0.1
                    ),
                    "verbose": -1,
                    "gpu_use_dp": False,
                }

                pruning_callback = optuna.integration.LightGBMPruningCallback(
                    trial, "binary_logloss"
                )
                result = lgb.cv(
                    param,
                    train_set=dtrain,
                    nfold=10,
                    num_boost_round=param["num_boost_round"],
                    early_stopping_rounds=10,
                    callbacks=[pruning_callback],
                    seed=42,
                    verbose_eval=False,
                )
                avg_result = np.mean(np.array(result["binary_logloss-mean"]))
                return avg_result

            algorithm = "lgbm"
            sampler = optuna.samplers.TPESampler(multivariate=True, seed=42)
            study = optuna.create_study(
                direction="minimize",
                sampler=sampler,
                study_name=f"{algorithm} tuning",
            )

            study.optimize(
                objective,
                n_trials=500,
                timeout=2 * 60 * 60,
                gc_after_trial=True,
                show_progress_bar=True,
            )

            lgbm_best_param = study.best_trial.params
            param = {
                "objective": "binary",
                "metric": "binary_logloss",
                # 'scale_pos_weight': lgbm_best_param["scale_pos_weight"],
                # 'max_depth': lgbm_best_param["max_depth"],
                "num_boost_round": lgbm_best_param["num_boost_round"],
                "lambda_l1": lgbm_best_param["lambda_l1"],
                "lambda_l2": lgbm_best_param["lambda_l2"],
                "num_leaves": lgbm_best_param["num_leaves"],
                "feature_fraction": lgbm_best_param["feature_fraction"],
                "feature_fraction_bynode": lgbm_best_param["feature_fraction_bynode"],
                # 'bagging_freq': lgbm_best_param["bagging_freq"],
                # 'min_child_samples': lgbm_best_param["min_child_samples"],
                "learning_rate": lgbm_best_param["learning_rate"],
                "verbose": -1,
                "gpu_use_dp": False,
            }
            Dtrain = lgb.Dataset(X_train, label=Y_train)
            Dtest = lgb.Dataset(X_test, label=Y_test)
            model = lgb.train(
                param,
                Dtrain,
                valid_sets=[Dtrain, Dtest],
                valid_names=["train", "valid"],
                early_stopping_rounds=10,
            )
            self.lgbm_2nd_meta_model = model
        else:
            nb_classes = len(self.unique_classes)

            def objective(trial):
                param = {
                    "objective": "multiclass",
                    "metric": "multi_logloss",
                    "boosting": "dart",
                    "drop_rate": trial.suggest_uniform("drop_rate", 0.1, 1.0),
                    "skip_drop": trial.suggest_uniform("skip_drop", 0.1, 1.0),
                    "num_boost_round": trial.suggest_int("num_boost_round", 100, 50000),
                    "num_class": nb_classes,
                    "lambda_l1": trial.suggest_loguniform("lambda_l1", 1, 1e6),
                    "lambda_l2": trial.suggest_loguniform("lambda_l2", 1, 1e6),
                    # 'max_depth': trial.suggest_int('max_depth', 2, 8), #-1
                    "num_leaves": trial.suggest_int("num_leaves", 2, 50),
                    "feature_fraction": trial.suggest_uniform(
                        "feature_fraction", 0.2, 1.0
                    ),
                    # 'bagging_freq': trial.suggest_int('bagging_freq', 1, 7),
                    "feature_fraction_bynode": trial.suggest_uniform(
                        "feature_fraction_bynode", 0.4, 1.0
                    ),
                    # 'min_child_samples': trial.suggest_int('min_child_samples', 5, 100),
                    "min_gain_to_split": trial.suggest_uniform(
                        "min_gain_to_split", 0, 15
                    ),
                    "learning_rate": trial.suggest_loguniform(
                        "learning_rate", 1e-5, 0.1
                    ),
                    "verbose": -1,
                    "gpu_use_dp": False,
                }

                pruning_callback = optuna.integration.LightGBMPruningCallback(
                    trial, "multi_logloss"
                )
                try:
                    result = lgb.cv(
                        param,
                        train_set=dtrain,
                        nfold=10,
                        num_boost_round=param["num_boost_round"],
                        early_stopping_rounds=10,
                        callbacks=[pruning_callback],
                        seed=42,
                        verbose_eval=False,
                    )
                    avg_result = np.mean(np.array(result["multi_logloss-mean"]))
                    # avg_result = self.meissner_cv_score(result["multi_logloss-mean"], penality_is_deducted=False)
                    return avg_result
                except Exception:
                    avg_result = 100
                return avg_result

            algorithm = "lgbm"
            sampler = optuna.samplers.TPESampler(multivariate=True, seed=42)
            study = optuna.create_study(
                direction="minimize",
                sampler=sampler,
                study_name=f"{algorithm} tuning",
            )
            study.optimize(
                objective,
                n_trials=500,
                timeout=2 * 60 * 60,
                gc_after_trial=True,
                show_progress_bar=True,
            )

            lgbm_best_param = study.best_trial.params
            param = {
                "objective": "multiclass",
                "metric": "multi_logloss",
                # 'class_weight': classes_weights,
                "boosting": "dart",
                "drop_rate": lgbm_best_param["drop_rate"],
                "skip_drop": lgbm_best_param["skip_drop"],
                "num_boost_round": lgbm_best_param["num_boost_round"],
                "num_class": nb_classes,
                "lambda_l1": lgbm_best_param["lambda_l1"],
                "lambda_l2": lgbm_best_param["lambda_l2"],
                # 'max_depth': lgbm_best_param["max_depth"],
                "num_leaves": lgbm_best_param["num_leaves"],
                "feature_fraction": lgbm_best_param["feature_fraction"],
                "feature_fraction_bynode": lgbm_best_param["feature_fraction_bynode"],
                # 'bagging_freq': lgbm_best_param["bagging_freq"],
                # 'min_child_samples': lgbm_best_param["min_child_samples"],
                "min_gain_to_split": lgbm_best_param["min_gain_to_split"],
                "learning_rate": lgbm_best_param["learning_rate"],
                "verbose": -1,
                "gpu_use_dp": False,
            }
            Dtrain = lgb.Dataset(X_train, label=Y_train)
            Dtest = lgb.Dataset(X_test, label=Y_test)
            model = lgb.train(
                param,
                Dtrain,
                valid_sets=[Dtrain, Dtest],
                valid_names=["train", "valid"],
                early_stopping_rounds=10,
            )
            self.lgbm_2nd_meta_model = model

    def predict(self, dataframe: pd.DataFrame, return_df=False):
        dataframe = self.predict_1st_layer(dataframe)
        model = self.lgbm_2nd_meta_model
        y_hat = model.predict(dataframe[self.selected_feats_2nd_layer])  # type:ignore
        self.lgbm_final_predictions = y_hat  # type:ignore
        if return_df:
            for lookup_class in range(self.unique_classes.shape[0]):
                dataframe[
                    f"lgbm_final_prediction_proba_{lookup_class}"
                ] = self.lgbm_final_predictions[
                    :, lookup_class
                ].reshape(  # type:ignore
                    -1, 1
                )  # type: ignore
            return dataframe
        else:
            return y_hat

    def predict_proba(self, dataframe: pd.DataFrame, return_df=False):
        dataframe = self.predict_1st_layer(dataframe)
        model = self.lgbm_2nd_meta_model
        y_hat = model.predict(dataframe[self.selected_feats_2nd_layer])  # type:ignore
        self.lgbm_final_predictions = y_hat
        if return_df:
            for lookup_class in range(self.unique_classes.shape[0]):
                dataframe[
                    f"lgbm_final_prediction_proba_{lookup_class}"
                ] = self.lgbm_final_predictions[
                    :, lookup_class
                ].reshape(  # type:ignore
                    -1, 1
                )  # type:ignore
            return dataframe
        else:
            return y_hat



import gc
import logging

import dill as pickle
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import shap
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    confusion_matrix,
    f1_score,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    median_absolute_error,
    r2_score,
    recall_score,
    roc_auc_score,
)

from e2eml.full_processing import cpu_preprocessing


def save_to_production(
    class_instance,
    file_path=None,
    file_name="automl_instance",
    file_type=".dat",
    clean=True,
):
    """
    Takes a pretrained model and saves it via pickle.
    :param class_instance: Takes instance of a class.
    :param file_name: Takes a string containing the whole file name.
    :param file_type: Takes the expected type of file to export.
    :return:
    """
    if clean:
        del class_instance.df_dict
        del class_instance.dataframe
        _ = gc.collect()
    else:
        pass

    if file_path:
        full_path = file_path + file_name + file_type
    else:
        full_path = file_name + file_type
    filehandler = open(full_path, "wb")
    pickle.dump(class_instance, filehandler)
    filehandler.close()


def load_for_production(file_path=None, file_name="automl_instance", file_type=".dat"):
    """
    Load in a pretrained auto ml model. This function will try to load the model as provided.
    It has a fallback logic to impute .dat as file_type in case the import fails initially.
    :param file_name:
    :param file_type:
    :return:
    """
    if file_path:
        full_path = file_path + file_name
    else:
        full_path = file_name
    try:
        filehandler = open(full_path, "rb")
    except Exception:
        filehandler = open(full_path + file_type, "rb")
    automl_model = pickle.load(filehandler)
    filehandler.close()
    return automl_model


class FullPipeline(cpu_preprocessing.PreProcessing):
    def shap_explanations(self, model, test_df, cols, explainer="tree", algorithm=None):
        """
        See explanations under:
        https://medium.com/rapids-ai/gpu-accelerated-shap-values-with-xgboost-1-3-and-rapids-587fad6822
        :param model: Trained ML model
        :param test_df: Test data to predict on.
        :param explainer: Set "tree" for TreeExplainer. Otherwise uses KernelExplainer.
        :param algorithm: Define name of the chosen ml algorithm as a string.
        :return: Returns plot of feature importance and interactions.
        """
        logging.info("Started creating SHAP values.")
        # print the JS visualization code to the notebook
        shap.initjs()
        if self.prediction_mode:
            to_pred = test_df
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            to_pred = X_test
        if explainer == "tree":
            try:
                explainer = shap.TreeExplainer(model)
                shap_values = explainer.shap_values(to_pred)
                shap.summary_plot(shap_values, to_pred, plot_type="bar", show=False)
                plt.savefig(f"{algorithm}Shap_feature_importance.png")
                plt.show()
            except AssertionError:
                model_shap_explainer = shap.KernelExplainer(model.predict, to_pred)
                model_shap_values = model_shap_explainer.shap_values(to_pred)
                shap.summary_plot(model_shap_values, to_pred, show=False)
                plt.savefig(f"{algorithm}Shap_feature_importance.png")
                plt.show()
        else:
            model_shap_explainer = shap.KernelExplainer(model.predict, to_pred)
            model_shap_values = model_shap_explainer.shap_values(to_pred)
            shap.summary_plot(model_shap_values, to_pred, show=False)
            plt.savefig(f"{algorithm}Shap_feature_importance.png")
            plt.show()
        logging.info("Finished creating SHAP values.")

    def visualize_probability_distribution(self, probabilities, threshold=0.5):
        """
        Vizualize the predicted probabilities in binary classification taks.
        :param probabilities: Numpy array or list of predicted probabilities.
        :param threshold: The threshold from where a probability shall counted towards the target class.
        :return: Displays a plotly line graph.
        """
        classes = [1 if prob > threshold else 0 for prob in probabilities]
        pred_dict = {"probabilities": probabilities, "classes": classes}
        pred_df = pd.DataFrame(pred_dict)
        pred_df["dummy_counter"] = 1
        pred_df = pred_df.round({"probabilities": 2})
        pred_df_gr = (
            pred_df.groupby(by=["classes", "probabilities"])
            .agg({"dummy_counter": "count"})
            .reset_index()
        )
        pred_df_gr = pred_df_gr.rename(columns={"dummy_counter": "nb_predictions"})
        fig = px.histogram(
            pred_df_gr,
            x="probabilities",
            y="nb_predictions",
            color="classes",
            nbins=10,
            title="Distribution of predicted probabailities and classes.",
        )
        fig.show()

    def target_size_estimator(self, probabilities, precision=0.50, recall=0.50):
        # TODO: Calculate Recall & Precision per class!!!!

        correct_found = []
        audience_sizes = []
        loop_spots = np.linspace(0, 1, 100, endpoint=False)
        for threshold in loop_spots:
            classes = [1 if prob > threshold else 0 for prob in probabilities]
            pred_dict = {"probabilities": probabilities, "classes": classes}
            pred_df = pd.DataFrame(pred_dict)
            pred_df["dummy_counter"] = 1

            pos_class_df = pred_df[(pred_df["classes"] == 1)].copy()
            total_pos = len(pos_class_df.index)
            estimated_pos_classes_found = total_pos * recall
            audience_size_needed = estimated_pos_classes_found / precision

            correct_found.append(estimated_pos_classes_found)
            audience_sizes.append(audience_size_needed)

    def classification_eval(  # noqa: C901
        self, algorithm, pred_probs=None, pred_class=None
    ):
        """
        Takes in the algorithm name. This is needed to grab saved predictions and to store cvlassification scores
        of different evaluation functions within the class. Returns the evaluation dictionary.
        :param algorithm: Name of the used algorithm
        :param pred_probs: Probabilities of predictions
        :param pred_class: Predicted classes
        :return: Returns the evaluation dictionary.
        """
        if self.prediction_mode:
            logging.info("Skipped classification evaluation due to prediction mode.")
            pass
        else:
            logging.info("Started classification evaluation.")
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            """
            We need a fallback logic as we might receive different types of data.
            If pred_class will not work with numpy arrays and needs the .any() function.
            """
            try:
                if pred_class:
                    y_hat = pred_class
                else:
                    y_hat = self.predicted_classes[f"{algorithm}"]
            except Exception:
                if pred_class.any():
                    y_hat = pred_class
                else:
                    y_hat = self.predicted_classes[f"{algorithm}"]

            try:
                if pred_probs:
                    y_hat_probs = pred_probs
                else:
                    y_hat_probs = self.predicted_probs[f"{algorithm}"]
            except Exception:
                try:
                    if pred_probs.any():
                        y_hat_probs = pred_probs
                    else:
                        y_hat_probs = self.predicted_probs[f"{algorithm}"]
                except Exception:
                    y_hat_probs = None

            """
            Calculating Matthews, ROC_AUC score and different F1 scores.
            """
            # y_hat = y_hat.astype(int)
            # Y_test = Y_test.astype(int)

            if self.class_problem == "binary":

                def get_preds(threshold, probabilities):
                    return [1 if prob > threshold else 0 for prob in probabilities]

                try:
                    if y_hat_probs:
                        roc_values = []
                        for thresh in np.linspace(0, 1, 100):
                            preds = get_preds(thresh, y_hat_probs)
                            tn, fp, fn, tp = confusion_matrix(Y_test, preds).ravel()
                            tpr = tp / (tp + fn)
                            fpr = fp / (fp + tn)
                            roc_values.append([tpr, fpr])
                        tpr_values, fpr_values = zip(*roc_values)
                        fig, ax = plt.subplots(figsize=(10, 7))
                        ax.plot(fpr_values, tpr_values)
                        ax.plot(
                            np.linspace(0, 1, 100),
                            np.linspace(0, 1, 100),
                            label="baseline",
                            linestyle="--",
                        )
                        plt.title(
                            f"Receiver Operating Characteristic Curve for {algorithm}",
                            fontsize=18,
                        )
                        plt.ylabel("TPR", fontsize=16)
                        plt.xlabel("FPR", fontsize=16)
                        plt.legend(fontsize=12)
                        plt.show()
                        roc_auc = roc_auc_score(Y_test, y_hat_probs)
                        print(f"The ROC_AUC score is {roc_auc}")

                        # plot predicted probabilities
                        self.visualize_probability_distribution(
                            y_hat_probs, threshold=0.5
                        )
                except ValueError:
                    if y_hat_probs.any():
                        roc_values = []
                        for thresh in np.linspace(0, 1, 100):
                            preds = get_preds(thresh, y_hat_probs)
                            tn, fp, fn, tp = confusion_matrix(Y_test, preds).ravel()
                            tpr = tp / (tp + fn)
                            fpr = fp / (fp + tn)
                            roc_values.append([tpr, fpr])
                        tpr_values, fpr_values = zip(*roc_values)
                        fig, ax = plt.subplots(figsize=(10, 7))
                        ax.plot(fpr_values, tpr_values)
                        ax.plot(
                            np.linspace(0, 1, 100),
                            np.linspace(0, 1, 100),
                            label="baseline",
                            linestyle="--",
                        )
                        plt.title(
                            f"Receiver Operating Characteristic Curve for {algorithm}",
                            fontsize=18,
                        )
                        plt.ylabel("TPR", fontsize=16)
                        plt.xlabel("FPR", fontsize=16)
                        plt.legend(fontsize=12)
                        plt.show()
                        roc_auc = roc_auc_score(Y_test, y_hat_probs)
                        print(f"The ROC_AUC score is {roc_auc}")

                        # plot predicted probabilities
                        self.visualize_probability_distribution(
                            y_hat_probs, threshold=0.5
                        )
                else:
                    roc_auc = 0.50
            else:
                roc_auc = 0.50
            try:
                matthews = matthews_corrcoef(Y_test, y_hat)
            except Exception:
                matthews = 0
            print(f"The Matthew correlation is {matthews}")
            logging.info(f"The Matthew correlation of {algorithm} is {matthews}")
            print("-------------------")
            accuracy = accuracy_score(Y_test, y_hat)
            print(f"The accuracy is {accuracy}")
            recall = recall_score(Y_test, y_hat, average="weighted")
            print(f"The recall is {recall}")
            f1_score_macro = f1_score(Y_test, y_hat, average="macro", zero_division=0)
            print(f"The macro F1 score is {f1_score_macro}")
            logging.info(f"The macro F1 score of {algorithm} is {f1_score_macro}")
            f1_score_micro = f1_score(Y_test, y_hat, average="micro", zero_division=0)
            print(f"The micro F1 score is {f1_score_micro}")
            logging.info(f"The micro F1 score of {algorithm} is {f1_score_micro}")
            f1_score_weighted = f1_score(
                Y_test, y_hat, average="weighted", zero_division=0
            )
            print(f"The weighted F1 score is {f1_score_weighted}")
            logging.info(f"The weighted F1 score of {algorithm} is {f1_score_weighted}")

            full_classification_report = classification_report(Y_test, y_hat)
            print(full_classification_report)

            logging.info(
                f"The classification report of {algorithm} is {full_classification_report}"
            )
            try:
                self.evaluation_scores[f"{algorithm}"] = {
                    "matthews": matthews,
                    "accuracy": accuracy,
                    "recall": recall,
                    "roc_auc": roc_auc,
                    "f1_score_macro": f1_score_macro,
                    "f1_score_micro": f1_score_micro,
                    "f1_score_weighted": f1_score_weighted,
                    "classfication_report": full_classification_report,
                }
            except UnboundLocalError:
                self.evaluation_scores[f"{algorithm}"] = {
                    "matthews": matthews,
                    "accuracy": accuracy,
                    "recall": recall,
                    "roc_auc": 0.50,
                    "f1_score_macro": f1_score_macro,
                    "f1_score_micro": f1_score_micro,
                    "f1_score_weighted": f1_score_weighted,
                    "classfication_report": full_classification_report,
                }
            logging.info("Finished classification evaluation.")
            return self.evaluation_scores

    def regression_eval(self, algorithm, pred_reg=None):
        """
        Takes in the algorithm name. This is needed to grab saved predictions and to store regression scores
        of different evaluation functions within the class. Returns the evaluation dictionary.
        :param algorithm: Name of the used algorithm
        :param pred_probs: Probabilities of predictions
        :param pred_class: Predicted classes
        :return: Returns the evaluation dictionary.
        """
        if self.prediction_mode:
            logging.info("Skipped regression evaluation due to prediction mode.")
            pass
        else:
            logging.info("Started regression evaluation.")
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            """
            We need a fallback logic as we might receive different types of data.
            If pred_class will not work with numpy arrays and needs the .any() function.
            """
            try:
                if pred_reg:
                    y_hat = pred_reg
                else:
                    y_hat = self.predicted_values[f"{algorithm}"]
            except Exception:
                if pred_reg.any():
                    y_hat = pred_reg
                else:
                    y_hat = self.predicted_values[f"{algorithm}"]

            """
            Calculating Matthews, ROC_AUC score and different F1 scores.
            """
            r2 = r2_score(Y_test, y_hat)
            print(f"The R2 score is {r2}")
            mean_absolute_error_score = mean_absolute_error(Y_test, y_hat)
            print(f"The MAE score is {mean_absolute_error_score}")
            median_absolute_error_score = median_absolute_error(Y_test, y_hat)
            print(f"The Median absolute error score is {median_absolute_error_score}")
            mean_squared_error_score = mean_squared_error(Y_test, y_hat, squared=False)
            print(f"The MSE score is {mean_squared_error_score}")
            root_mean_squared_error_score = mean_squared_error(
                Y_test, y_hat, squared=True
            )
            print(f"The RMSE score is {root_mean_squared_error_score}")

            self.evaluation_scores[f"{algorithm}"] = {
                "mae": mean_absolute_error_score,
                "r2_score": r2,
                "MSE": mean_squared_error_score,
                "RMSE": root_mean_squared_error_score,
                "median_absolute_error": median_absolute_error_score,
            }
            logging.info("Finished regression evaluation.")
            return self.evaluation_scores



import gc
import logging
import random
import re
import ssl
import string

import nltk
import numpy as np
import pandas as pd
import psutil
import spacy
import transformers
from imblearn.over_sampling import SMOTE
from nltk import pos_tag
from nltk.corpus import wordnet
from nltk.tokenize import word_tokenize
from sklearn.decomposition import PCA
from sklearn.feature_extraction.text import TfidfVectorizer
from textblob import TextBlob
from transformers import AutoConfig, AutoModel, AutoTokenizer

from e2eml.full_processing import cpu_preprocessing

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    print("Created unverified SSL context to be able to download NLTK dictionaries.")
else:
    ssl._create_default_https_context = _create_unverified_https_context
nltk.download("punkt")
nltk.download("averaged_perceptron_tagger")
nltk.download("stopwords")
nltk.download("wordnet")
# stop_words = set(stopwords.words(self.blueprint_step_selection_nlp_transformers["synonym_language"]))

# TODO: Continue on NLP
"""
THIS PART IS UNFINISHED AND UNTESTED.
"""


class NlpPreprocessing(cpu_preprocessing.PreProcessing):
    """
    Preprocess a string.
    :parameter
        :param text: string - name of column containing text
        :param lst_stopwords: list - list of stopwords to remove
        :param flg_stemm: bool - whether stemming is to be applied
        :param flg_lemm: bool - whether lemmitisation is to be applied
    :return
        cleaned text
    """

    def utils_preprocess_text(
        self, text, flg_stemm=False, flg_lemm=True, lst_stopwords=None
    ):
        logging.info("Start text cleaning.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        # clean (convert to lowercase and remove punctuations and   characters and then strip)
        text = re.sub(r"[^\w\s]", "", str(text).lower().strip())

        # Tokenize (convert from string to list)
        lst_text = text.split()  # remove Stopwords
        if lst_stopwords is not None:
            lst_text = [word for word in lst_text if word not in lst_stopwords]

        # Stemming (remove -ing, -ly, ...)
        if flg_stemm is True:
            logging.info("Start text stemming.")
            ps = nltk.stem.porter.PorterStemmer()
            lst_text = [ps.stem(word) for word in lst_text]

        # Lemmatisation (convert the word into root word)
        if flg_lemm is True:
            logging.info("Start text lemmatisation.")
            lem = nltk.stem.wordnet.WordNetLemmatizer()
            lst_text = [lem.lemmatize(word) for word in lst_text]

        # back to string from list
        text = " ".join(lst_text)
        return text

    def remove_url(self, text):
        logging.info("Start removing URLs.")
        url = re.compile(r"https?://\S+|www\.\S+")
        return url.sub(r"", str(text))

    def remove_punct(self, text):
        logging.info("Start removing punctuation.")
        table = str.maketrans("", "", string.punctuation)
        return text.translate(table)

    def remove_html(self, text):
        logging.info("Start removing HTML.")
        html = re.compile(r"<.*?>")
        return html.sub(r"", str(text))

    def remove_emoji(self, text):
        logging.info("Start removing emojis.")
        emoji_pattern = re.compile(
            "["
            "\U0001F600-\U0001F64F"  # emoticons
            "\U0001F300-\U0001F5FF"  # symbols & pictographs
            "\U0001F680-\U0001F6FF"  # transport & map symbols
            "\U0001F1E0-\U0001F1FF"  # flags (iOS)
            "\U00002702-\U000027B0"
            "\U000024C2-\U0001F251"
            "]+",
            flags=re.UNICODE,
        )
        return emoji_pattern.sub(r"", str(text))

    def decontraction(self, text):
        logging.info("Start decontraction.")
        text = re.sub(r"won\'t", " will not", text)
        text = re.sub(r"won\'t've", " will not have", text)
        text = re.sub(r"can\'t", " can not", text)
        text = re.sub(r"don\'t", " do not", text)

        text = re.sub(r"can\'t've", " can not have", text)
        text = re.sub(r"ma\'am", " madam", text)
        text = re.sub(r"let\'s", " let us", text)
        text = re.sub(r"ain\'t", " am not", text)
        text = re.sub(r"shan\'t", " shall not", text)
        text = re.sub(r"sha\n't", " shall not", text)
        text = re.sub(r"o\'clock", " of the clock", text)
        text = re.sub(r"y\'all", " you all", text)

        text = re.sub(r"n\'t", " not", text)
        text = re.sub(r"n\'t've", " not have", text)
        text = re.sub(r"\'re", " are", text)
        text = re.sub(r"\'s", " is", text)
        text = re.sub(r"\'d", " would", text)
        text = re.sub(r"\'d've", " would have", text)
        text = re.sub(r"\'ll", " will", text)
        text = re.sub(r"\'ll've", " will have", text)
        text = re.sub(r"\'t", " not", text)
        text = re.sub(r"\'ve", " have", text)
        text = re.sub(r"\'m", " am", text)
        text = re.sub(r"\'re", " are", text)
        return str(text)

    def seperate_alphanumeric(self, text):
        logging.info("Start seperating alphanumeric.")
        words = text
        words = re.findall(r"[^\W\d_]+|\d+", words)
        return " ".join(words)

    def cont_rep_char(self, text):
        tchr = text.group(0)

        if len(tchr) > 1:
            return tchr[0:2]

    def unique_char(self, rep, text):
        substitute = re.sub(r"(\w)\1+", rep, text)
        return str(substitute)

    def regex_clean_text_data(self):  # noqa: C901
        logging.info("Start text cleaning.")
        self.get_current_timestamp(task="Start text cleaning.")
        logging.info("Start text cleaning.")
        if self.prediction_mode:
            text_columns = []
            text_columns.append(self.nlp_transformer_columns)
            for text_col in text_columns:
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.utils_preprocess_text(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.remove_url(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.remove_punct(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.remove_emoji(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.decontraction(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.seperate_alphanumeric(x)
                )
                self.dataframe[text_col] = self.dataframe[text_col].apply(
                    lambda x: self.unique_char(self.cont_rep_char, x)
                )
            logging.info("Finished text cleaning.")
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            text_columns = []
            text_columns.append(self.nlp_transformer_columns)

            for text_col in text_columns:
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.utils_preprocess_text(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.remove_url(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.remove_punct(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.remove_emoji(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.decontraction(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.seperate_alphanumeric(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(
                        lambda x: self.unique_char(self.cont_rep_char, x)
                    )
                except (TypeError, AttributeError):
                    pass

                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.utils_preprocess_text(x)
                    )
                except (TypeError, AttributeError):
                    pass

                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.remove_url(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.remove_punct(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.remove_emoji(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.decontraction(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.seperate_alphanumeric(x)
                    )
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(
                        lambda x: self.unique_char(self.cont_rep_char, x)
                    )
                except (TypeError, AttributeError):
                    pass
            logging.info("Finished text cleaning.")
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def textBlob_sentiment_polarity_score(self, text):
        # This polarity score is between -1 to 1
        polarity = TextBlob(text).sentiment.polarity
        return polarity

    def append_text_sentiment_score(self, text_columns=None):
        self.get_current_timestamp(task="Start text cleaning.")
        logging.info("Start sentiment polarity score.")
        algorithm = "textblob_sentiment_score"
        if self.prediction_mode:
            text_columns = self.nlp_transformer_columns
            for text_col in [text_columns]:
                self.dataframe[f"{algorithm}_{text_col}"] = self.dataframe[
                    text_col
                ].apply(lambda x: self.textBlob_sentiment_polarity_score(x))
            logging.info("Finished text cleaning.")
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            text_columns = self.nlp_transformer_columns
            for text_col in [text_columns]:
                X_train[f"{algorithm}_{text_col}"] = X_train[text_col].apply(
                    lambda x: self.textBlob_sentiment_polarity_score(x)
                )
                X_test[f"{algorithm}_{text_col}"] = X_test[text_col].apply(
                    lambda x: self.textBlob_sentiment_polarity_score(x)
                )
            logging.info("Finished text cleaning.")
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def spacy_features(self, df: pd.DataFrame, text_column):
        """
        This function generates features using spacy en_core_wb_lg
        More information:
        https://www.kaggle.com/konradb/linear-baseline-with-cv
        https://www.kaggle.com/anaverageengineer/comlrp-baseline-for-complete-beginners

        """
        logging.info("Download spacy language package.")
        try:
            nlp = spacy.load("en_core_web_sm")
        except OSError:
            print(
                "Downloading language model for the spaCy POS tagger\n"
                "(don't worry, this will only happen once)"
            )
            from spacy.cli import download

            download("en_core_web_sm")
            nlp = spacy.load("en_core_web_sm")
            # https://spacy.io/universe/project/spacy-transformers
            # https://spacy.io/models
            # !python -m spacy download en_core_web_trf
            # !pip install spacy-transformers
            # nlp = spacy.load('en_core_web_trf')
        # nlp = spacy.load('en_core_web_lg')
        with nlp.disable_pipes():
            vectors = [nlp(text).vector for text in df[text_column]]
        return vectors

    def get_spacy_col_names(self):
        logging.info("Get spacy column names.")
        names = list()
        for i in range(96):  # TODO: Why 96?
            names.append(f"spacy_{i}")
        return names

    def pos_tag_features(self, passage: str):
        """
        This function counts the number of times different parts of speech occur in an excerpt. POS tags are listed here:
        https://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html
        """
        logging.info("Get POS tags.")
        pos_tags = [
            "CC",
            "CD",
            "DT",
            "EX",
            "FW",
            "IN",
            "JJ",
            "JJR",
            "JJS",
            "LS",
            "MD",
            "NN",
            "NNS",
            "NNP",
            "NNPS",
            "PDT",
            "POS",
            "PRP",
            "RB",
            "RBR",
            "RBS",
            "RP",
            "TO",
            "UH",
            "VB",
            "VBD",
            "VBG",
            "VBZ",
            "WDT",
            "WP",
            "WRB",
        ]

        tags = pos_tag(word_tokenize(passage))
        tag_list = list()
        for tag in pos_tags:
            tag_list.append(len([i[0] for i in tags if i[1] == tag]))
        return tag_list

    def pos_tagging_pca_nlp_cols(self, df, text_cols, mode="fit", pca_pos_tags=True):
        if mode == "transform":
            pass
        else:
            pos_columns = set([])

        if self.nlp_columns:
            text_cols = self.nlp_columns
        else:
            pass

        target_columns = [
            "CC",
            "CD",
            "DT",
            "EX",
            "FW",
            "IN",
            "JJ",
            "JJR",
            "JJS",
            "LS",
            "MD",
            "NN",
            "NNS",
            "NNP",
            "NNPS",
            "PDT",
            "POS",
            "PRP",
            "RB",
            "RBR",
            "RBS",
            "RP",
            "TO",
            "UH",
            "VB",
            "VBD",
            "VBG",
            "VBZ",
            "WDT",
            "WP",
            "WRB",
        ]
        if mode == "transform":
            for text_col in text_cols:
                temp_target_columns = [x + text_col for x in target_columns]
                df[text_col].fillna("None", inplace=True)
                spacy_df = pd.DataFrame(
                    self.spacy_features(df, text_col),
                    columns=self.get_spacy_col_names(),
                )
                temp_df = pd.merge(
                    df, spacy_df, left_index=True, right_index=True, how="left"
                )
                pos_df = pd.DataFrame(
                    temp_df[text_col]
                    .apply(lambda p: self.pos_tag_features(p))
                    .tolist(),
                    columns=temp_target_columns,
                )
                pos_df.fillna(0, inplace=True)
                if pca_pos_tags:
                    self.get_current_timestamp(task="PCA POS tags")
                    logging.info("Start to PCA POS tags.")
                    pca = self.preprocess_decisions["spacy_pos"][f"pos_pca_{text_col}"]
                    comps = pca.transform(pos_df)
                    pos_pca_cols = [f"POS PC-1 {text_col}", f"POS PC-2 {text_col}"]
                    pos_df = pd.DataFrame(comps, columns=pos_pca_cols)
                    pos_df_pca = pos_df[pos_pca_cols]
                    df = pd.merge(
                        df, pos_df_pca, left_index=True, right_index=True, how="left"
                    )
                else:
                    df = pd.merge(
                        df, pos_df, left_index=True, right_index=True, how="left"
                    )
        elif mode == "fit":

            # if self.nlp_columns
            nlp_columns = []
            for text_col in text_cols:
                try:
                    # do we have at least 3 words?
                    df[f"nof_words_{text_col}"] = df[text_col].apply(
                        lambda s: len(s.split(" "))
                    )
                    if df[f"nof_words_{text_col}"].max() >= 3:
                        temp_target_columns = [x + text_col for x in target_columns]
                        df[text_col].fillna("None", inplace=True)
                        spacy_df = pd.DataFrame(
                            self.spacy_features(df, text_col),
                            columns=self.get_spacy_col_names(),
                        )
                        temp_df = pd.merge(
                            df, spacy_df, left_index=True, right_index=True, how="left"
                        )
                        pos_df = pd.DataFrame(
                            temp_df[text_col]
                            .apply(lambda p: self.pos_tag_features(p))
                            .tolist(),
                            columns=temp_target_columns,
                        )
                        pos_df.fillna(0, inplace=True)
                        for col in pos_df.columns:
                            pos_columns.add(col)
                        if pca_pos_tags:
                            self.get_current_timestamp(task="PCA POS tags")
                            logging.info("Start to PCA POS tags.")
                            pca = PCA(n_components=2)
                            comps = pca.fit_transform(pos_df)
                            self.preprocess_decisions["spacy_pos"][
                                f"pos_pca_{text_col}"
                            ] = pca
                            pos_pca_cols = [
                                f"POS PC-1 {text_col}",
                                f"POS PC-2 {text_col}",
                            ]
                            pos_df = pd.DataFrame(comps, columns=pos_pca_cols)
                            pos_df_pca = pos_df[pos_pca_cols]
                            df = pd.merge(
                                df,
                                pos_df_pca,
                                left_index=True,
                                right_index=True,
                                how="left",
                            )
                        else:
                            df = pd.merge(
                                df,
                                pos_df,
                                left_index=True,
                                right_index=True,
                                how="left",
                            )
                        nlp_columns.append(text_col)
                    else:
                        pass
                    df.drop(f"nof_words_{text_col}", axis=1, inplace=True)
                except AttributeError:
                    pass
            if pca_pos_tags:
                # get unique pos tag columns
                unique_pos_cols = list(set(pos_columns))
                self.preprocess_decisions["spacy_pos"][
                    "pos_tagger_cols"
                ] = unique_pos_cols

            if self.nlp_columns:
                pass
            else:
                # get unique original column names
                unique_nlp_cols = list(set(nlp_columns))
                self.nlp_columns = unique_nlp_cols
        try:
            del pca
            del spacy_df
            del temp_df
            del comps
            _ = gc.collect()
        except Exception:
            pass
        return df

    def pos_tagging_pca(self, pca_pos_tags=True):
        self.get_current_timestamp(task="Start Spacy, POS tagging")
        logging.info("Start spacy POS tagging loop.")
        # algorithm = "spacy_pos"
        if self.prediction_mode:
            text_columns = self.nlp_columns
            self.dataframe = self.pos_tagging_pca_nlp_cols(
                self.dataframe,
                text_columns,
                mode="transform",
                pca_pos_tags=pca_pos_tags,
            )
            logging.info("Finished spacy POS tagging loop.")
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            if pca_pos_tags:
                self.preprocess_decisions["spacy_pos"] = {}
            else:
                pass
            if not self.nlp_columns:
                text_columns = X_train.select_dtypes(include=["object"]).columns
            else:
                text_columns = self.nlp_columns
            X_train = self.pos_tagging_pca_nlp_cols(
                X_train, text_columns, mode="fit", pca_pos_tags=pca_pos_tags
            )
            X_test = self.pos_tagging_pca_nlp_cols(
                X_test, text_columns, mode="transform", pca_pos_tags=pca_pos_tags
            )
            logging.info("Finished spacy POS tagging loop.")
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def tfidf_pca(
        self, df, text_cols, mode="fit", pca_pos_tags=True, ngram_range=(1, 2)
    ):
        if self.nlp_columns:
            text_cols = self.nlp_columns
        else:
            pass
        if mode == "transform":
            for text_col in text_cols:
                df[text_col].fillna("None", inplace=True)
                try:
                    tfids = self.preprocess_decisions["tfidf_vectorizer"][
                        f"tfidf_{text_col}"
                    ]
                    if pca_pos_tags:
                        vector = list(tfids.transform(df[text_col]).toarray())
                        self.get_current_timestamp(task="PCA POS tags")
                        logging.info("Start to PCA POS tags.")
                        pca = self.preprocess_decisions["tfidf_vectorizer"][
                            f"tfidf_pca_{text_col}"
                        ]
                        comps = pca.transform(vector)
                        tfidf_pca_cols = [
                            f"TFIDF PC-1 {text_col}",
                            f"TFIDF PC-2 {text_col}",
                        ]
                        pos_df = pd.DataFrame(comps, columns=tfidf_pca_cols)
                        tfidf_df_pca = pos_df[tfidf_pca_cols]
                        df = pd.merge(
                            df,
                            tfidf_df_pca,
                            left_index=True,
                            right_index=True,
                            how="left",
                        )
                    else:
                        all_embeddings = tfids.transform(df[text_col]).toarray()
                        tfidf_df = pd.DataFrame(
                            all_embeddings, columns=tfids.get_feature_names()
                        )
                        tfidf_df = tfidf_df.add_prefix("tfids_")
                        df = pd.concat([df, tfidf_df], axis=1)
                except KeyError:
                    pass
        elif mode == "fit":
            if self.nlp_columns:
                text_cols = self.nlp_columns
            else:
                pass
            # if self.nlp_columnss
            nlp_columns = []
            for text_col in text_cols:
                try:
                    # do we have at least 1 word?
                    df[f"nof_words_{text_col}"] = df[text_col].apply(
                        lambda s: len(s.split(" "))
                    )
                    if df[f"nof_words_{text_col}"].max() >= 3:
                        df[text_col].fillna("None", inplace=True)
                        tfids = TfidfVectorizer(
                            ngram_range=ngram_range,
                            strip_accents="unicode",
                            max_features=25000,
                            max_df=0.85,
                            min_df=3,
                            norm="l2",
                            sublinear_tf=True,
                        )
                        self.preprocess_decisions["tfidf_vectorizer"][
                            f"tfidf_{text_col}"
                        ] = tfids
                        if pca_pos_tags:
                            vector = list(tfids.fit_transform(df[text_col]).toarray())
                            self.get_current_timestamp(task="PCA TfIDF matrix")
                            logging.info("Start to PCA TfIDF matrix.")
                            pca = PCA(n_components=2)
                            comps = pca.fit_transform(vector)
                            self.preprocess_decisions["tfidf_vectorizer"][
                                f"tfidf_pca_{text_col}"
                            ] = pca
                            tfidf_pca_cols = [
                                f"TFIDF PC-1 {text_col}",
                                f"TFIDF PC-2 {text_col}",
                            ]
                            pos_df = pd.DataFrame(comps, columns=tfidf_pca_cols)
                            tfidf_df_pca = pos_df[tfidf_pca_cols]
                            df = pd.merge(
                                df,
                                tfidf_df_pca,
                                left_index=True,
                                right_index=True,
                                how="left",
                            )
                        else:
                            all_embeddings = tfids.fit_transform(df[text_col]).toarray()
                            tfidf_df = pd.DataFrame(
                                all_embeddings, columns=tfids.get_feature_names()
                            )
                            tfidf_df = tfidf_df.add_prefix("tfids_")
                            df = pd.concat([df, tfidf_df], axis=1)
                        nlp_columns.append(text_col)
                    else:
                        pass
                    df.drop(f"nof_words_{text_col}", axis=1, inplace=True)
                except AttributeError:
                    pass
            if self.nlp_columns:
                pass
            else:
                # get unique original column names
                unique_nlp_cols = list(set(nlp_columns))
                self.nlp_columns = unique_nlp_cols
        try:
            del tfidf_df
            del all_embeddings
            del tfids
            del tfidf_df_pca
            _ = gc.collect()
        except Exception:
            pass
        return df

    def tfidf_vectorizer_to_pca(self, pca_pos_tags=True, ngram_range=(1, 3)):
        self.get_current_timestamp(task="Start TFIDF to PCA loop")
        logging.info("Start TFIDF to PCA loop.")
        if self.prediction_mode:
            text_columns = self.nlp_columns
            self.dataframe = self.tfidf_pca(
                self.dataframe,
                text_columns,
                mode="transform",
                pca_pos_tags=pca_pos_tags,
                ngram_range=ngram_range,
            )
            logging.info("Finished TFIDF to PCA loop.")
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            if pca_pos_tags:
                self.preprocess_decisions["tfidf_vectorizer"] = {}
            else:
                pass
            if not self.nlp_columns:
                text_columns = X_train.select_dtypes(include=["object"]).columns
            else:
                text_columns = self.nlp_columns
            X_train = self.tfidf_pca(
                X_train,
                text_columns,
                mode="fit",
                pca_pos_tags=pca_pos_tags,
                ngram_range=ngram_range,
            )
            X_test = self.tfidf_pca(
                X_test,
                text_columns,
                mode="transform",
                pca_pos_tags=pca_pos_tags,
                ngram_range=ngram_range,
            )
            logging.info("Finished TFIDF to PCA loop.")
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def tfidf_smote_text_columns(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            tfids = TfidfVectorizer(
                strip_accents="unicode",
                max_features=25000,
                max_df=0.85,
                min_df=3,
                norm="l2",
                sublinear_tf=True,
            )
            tfids.fit(X_train[self.nlp_transformer_columns])
            vectors = tfids.transform(X_train[self.nlp_transformer_columns])
            vectors_df = pd.DataFrame(
                vectors.todense(), columns=tfids.get_feature_names()
            )
            smt = SMOTE(random_state=777, k_neighbors=1)
            X_SMOTE, y_SMOTE = smt.fit_sample(vectors_df, Y_train)
            X_train = pd.DataFrame(X_SMOTE.todense(), columns=tfids.get_feature_names())
            print(X_train)
            Y_train = y_SMOTE
            del vectors
            del vectors_df
            del smt
            _ = gc.collect()
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def get_synonyms(self, word):
        """
        Get synonyms of a word
        """
        synonyms = set()

        for syn in wordnet.synsets(word):
            for lemma in syn.lemmas():
                synonym = lemma.name().replace("_", " ").replace("-", " ").lower()
                synonym = "".join(
                    [char for char in synonym if char in " qwertyuiopasdfghjklzxcvbnm"]
                )
                synonyms.add(synonym)

        if word in synonyms:
            synonyms.remove(word)

        return list(synonyms)

    def synonym_replacement(self, words, n):
        words = words.split()
        new_words = words.copy()
        stop_words = nltk.corpus.stopwords.words(
            self.blueprint_step_selection_nlp_transformers["synonym_language"]
        )
        random_word_list = list(set([word for word in words if word not in stop_words]))
        random.shuffle(random_word_list)
        num_replaced = 0

        for random_word in random_word_list:
            synonyms = self.get_synonyms(random_word)

            if len(synonyms) >= 1:
                synonym = random.choice(list(synonyms))
                new_words = [
                    synonym if word == random_word else word for word in new_words
                ]
                num_replaced += 1

            if num_replaced >= n:  # only replace up to n words
                break
        sentence = " ".join(new_words)
        return sentence

    def replace_synonyms_to_df_copy(self, words_to_replace=3, mode="auto"):
        """
        The function copies the original dataframe and randomly exchanges words by their synonyms. It concats the original to the
        modified dataframe and returns the new training dataset.
        :param words_to_replace:
        :return:
        """
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            X_train[self.target_variable] = Y_train
            if mode == "auto":
                sentence_length = X_train[self.nlp_transformer_columns].apply(
                    lambda x: np.max([len(w) for w in x.split()])
                )
                words_to_replace = int(
                    sentence_length.max() / 15
                )  # exchange fifteen percent of the words
            train_copy = X_train.copy()
            train_copy[self.nlp_transformer_columns] = train_copy[
                self.nlp_transformer_columns
            ].apply(lambda x: self.synonym_replacement(x, words_to_replace))
            X_train = pd.concat([X_train, train_copy], ignore_index=True)
            X_train = X_train.reset_index(drop=True)
            Y_train = X_train[self.target_variable]
            X_train.drop(self.target_variable, axis=1)

            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def create_bert_classification_model(self, chosen_model="bert-base-uncased"):
        logging.info("Start creating or loading transformer model for classification.")
        if not self.transformer_chosen:
            chosen_model = chosen_model
        config = AutoConfig.from_pretrained(
            self.transformer_chosen, num_labels=self.num_classes
        )
        model = transformers.AutoModelForSequenceClassification.from_pretrained(
            self.transformer_chosen, config=config
        )
        return model

    def create_bert_regression_model(self, chosen_model="bert-base-uncased"):
        logging.info("Start creating or loading transformer model for regression.")
        if not self.transformer_chosen:
            chosen_model = chosen_model
        config = AutoConfig.from_pretrained(self.transformer_chosen, num_labels=1)
        model = transformers.AutoModelForSequenceClassification.from_pretrained(
            self.transformer_chosen, config=config
        )
        return model

    def import_transformer_model_tokenizer(self, transformer_chosen=None):
        logging.info("Start importing transformer tokenizer.")
        if not transformer_chosen:
            transformer_chosen = "bert-base-uncased"
        else:
            transformer_chosen = self.transformer_chosen

        if self.transformer_model_load_from_path:
            bert = AutoModel.from_pretrained(
                f"{self.transformer_model_load_from_path}",
                output_attentions=False,  # Whether the model returns attentions weights.
                output_hidden_states=False,
            )
            tokenizer = transformers.AutoTokenizer.from_pretrained(
                f"{self.transformer_model_load_from_path}"
            )
        else:
            if self.class_problem in ["binary", "multiclass"]:
                # import BERT-base pretrained model
                bert = self.create_bert_classification_model(transformer_chosen)
            elif self.class_problem == "regression":
                bert = self.create_bert_regression_model(transformer_chosen)
            else:
                print("No correct ml_task defined during class instantiation.")
            # Load the BERT tokenizer
            tokenizer = AutoTokenizer.from_pretrained(transformer_chosen)
        if "nlp_transformers" in self.preprocess_decisions:
            pass
        else:
            self.preprocess_decisions["nlp_transformers"] = {}

        self.preprocess_decisions["nlp_transformers"][
            f"transformer_model_{transformer_chosen}"
        ] = bert
        self.preprocess_decisions["nlp_transformers"][
            f"transformer_tokenizer_{transformer_chosen}"
        ] = tokenizer



import re

import pandas as pd
from sklearn.metrics import classification_report, matthews_corrcoef
from sklearn.utils import shuffle

from e2eml.classification import classification_blueprints as cb

# import pytest


def load_titanic_data():
    """
    Load & preprocess Titanic dataset. The feature engineering simulates the business knowledge part.
    The code has been taken from:
    https://towardsdatascience.com/predicting-the-survival-of-titanic-passengers-30870ccc7e8
    :return: Several dataframes and series to be processed by blueprint.
    """
    data = pd.read_csv("titanic_train.csv")
    print("Create additional features and modify existing ones.")
    deck = {"A": 1, "B": 2, "C": 3, "D": 4, "E": 5, "": 6, "G": 7, "U": 8}
    data["Cabin"] = data["Cabin"].fillna("U0")
    data["Deck"] = data["Cabin"].map(
        lambda x: re.compile("([a-zA-Z]+)").search(x).group()
    )
    data["Deck"] = data["Deck"].map(deck)
    data["Deck"] = data["Deck"].fillna(0)
    data["Deck"] = data["Deck"].astype(int)

    data.loc[data["Fare"] <= 7.91, "Fare"] = 0
    data.loc[(data["Fare"] > 7.91) & (data["Fare"] <= 14.454), "Fare"] = 1
    data.loc[(data["Fare"] > 14.454) & (data["Fare"] <= 31), "Fare"] = 2
    data.loc[(data["Fare"] > 31) & (data["Fare"] <= 99), "Fare"] = 3
    data.loc[(data["Fare"] > 99) & (data["Fare"] <= 250), "Fare"] = 4
    data.loc[data["Fare"] > 250, "Fare"] = 5

    data["Age"].fillna(0, inplace=True)
    data["Age"] = data["Age"].astype(int)
    data.loc[data["Age"] <= 11, "Age"] = 0
    data.loc[(data["Age"] > 11) & (data["Age"] <= 18), "Age"] = 1
    data.loc[(data["Age"] > 18) & (data["Age"] <= 22), "Age"] = 2
    data.loc[(data["Age"] > 22) & (data["Age"] <= 27), "Age"] = 3
    data.loc[(data["Age"] > 27) & (data["Age"] <= 33), "Age"] = 4
    data.loc[(data["Age"] > 33) & (data["Age"] <= 40), "Age"] = 5
    data.loc[(data["Age"] > 40) & (data["Age"] <= 66), "Age"] = 6
    data.loc[data["Age"] > 66, "Age"] = 6

    titles = {"Mr": 1, "Miss": 2, "Mrs": 3, "Master": 4, "Rare": 5}
    # extract titles
    data["Title"] = data.Name.str.extract(r" ([A-Za-z]+)\.", expand=False)
    # replace titles with a more common title or as Rare
    data["Title"] = data["Title"].replace(
        [
            "Lady",
            "Countess",
            "Capt",
            "Col",
            "Don",
            "Dr",
            "Major",
            "Rev",
            "Sir",
            "Jonkheer",
            "Dona",
        ],
        "Rare",
    )
    data["Title"] = data["Title"].replace("Mlle", "Miss")
    data["Title"] = data["Title"].replace("Ms", "Miss")
    data["Title"] = data["Title"].replace("Mme", "Mrs")
    # convert titles into numbers
    data["Title"] = data["Title"].map(titles)
    # filling NaN with 0, to get safe
    data["Title"] = data["Title"].fillna(0)

    data["relatives"] = data["SibSp"] + data["Parch"]
    data.loc[data["relatives"] > 0, "not_alone"] = 0
    data.loc[data["relatives"] == 0, "not_alone"] = 1
    data["not_alone"] = data["not_alone"].astype(int)

    data["Fare"] = data["Fare"].astype(int)
    data["Age_Class"] = data["Age"] * data["Pclass"]
    data["Fare_Per_Person"] = data["Fare"] / (data["relatives"] + 1)
    data["Fare_Per_Person"] = data["Fare_Per_Person"].astype(int)
    print("Do dataframe splits.")
    test_df = data.head(800).copy()
    val_df = data.tail(91).copy()
    val_df_target = val_df["Survived"].copy()
    del val_df["Survived"]
    test_target = "Survived"
    test_categorical_cols = ["Pclass", "Name", "Sex", "PassengerId"]
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def steel_fault_multiclass_data():
    data = pd.read_csv("faults.csv")
    data = shuffle(data)
    # data["leakage"] = data["target"]
    print(data.info())
    # from sklearn.preprocessing import LabelEncoder
    # le = LabelEncoder()
    # X = data.drop('target', axis=1)
    # Y = le.fit_transform(data['target'])

    # X['target'] = Y
    test_df = data.head(1500).copy()
    val_df = data.tail(441).copy()
    val_df_target = val_df["target"].copy()
    del val_df["target"]
    test_target = "target"
    test_categorical_cols = None
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def nlp_multiclass_data():
    data = pd.read_csv("Corona_NLP_train.csv", encoding="latin-1")
    test_df = data.head(100).copy()
    print(test_df[["OriginalTweet"]])
    val_df = data.tail(499).copy()
    val_df_target = val_df["Sentiment"].copy()
    del val_df["Sentiment"]
    test_target = "Sentiment"
    test_categorical_cols = ["Location", "OriginalTweet"]
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def test_ml_special_multiclass_full_processing_multimodel_max_voting(dataset="titanic"):
    if dataset == "titanic":
        (
            test_df,
            test_target,
            val_df,
            val_df_target,
            test_categorical_cols,
        ) = load_titanic_data()
        titanic_auto_ml = cb.ClassificationBluePrint(
            datasource=test_df,
            target_variable=test_target,
            categorical_columns=test_categorical_cols,
            preferred_training_mode="auto",
            tune_mode="accurate",
            rapids_acceleration=True,
        )
        titanic_auto_ml.hyperparameter_tuning_sample_size = 1100

    elif dataset == "synthetic_multiclass":
        (
            test_df,
            test_target,
            val_df,
            val_df_target,
            test_categorical_cols,
        ) = steel_fault_multiclass_data()
        titanic_auto_ml = cb.ClassificationBluePrint(
            datasource=test_df,
            target_variable=test_target,
            categorical_columns=test_categorical_cols,
        )
    elif dataset == "corona_tweet":
        (
            test_df,
            test_target,
            val_df,
            val_df_target,
            test_categorical_cols,
        ) = nlp_multiclass_data()
        titanic_auto_ml = cb.ClassificationBluePrint(
            datasource=test_df,
            target_variable=test_target,
            categorical_columns=test_categorical_cols,
            preferred_training_mode="auto",
            tune_mode="accurate",
            nlp_transformer_columns="OriginalTweet",
        )

    titanic_auto_ml.hyperparameter_tuning_rounds = {
        "xgboost": 10,
        "lgbm": 15,
        "tabnet": 3,
        "ngboost": 10,
        "sklearn_ensemble": 3,
        "catboost": 10,
        "ridge": 25,
        "bruteforce_random": 10,
        "autoencoder_based_oversampling": 20,
        "autoencoder_based_dimensionality_reduction": 100,
        "final_pca_dimensionality_reduction": 20,
        "synthetic_data_augmentation": 100,
    }

    titanic_auto_ml.special_blueprint_algorithms = {
        "ridge": False,  # titanic, #synthetic_multiclass
        "xgboost": False,  # titanic
        "ngboost": False,  # titanic, #synthetic_multiclass
        "lgbm": True,  # titanic
        "tabnet": False,  # titanic
        "vowpal_wabbit": False,
        "sklearn_ensemble": False,  # titanic, #synthetic_multiclass
        "catboost": False,
    }
    titanic_auto_ml.blueprint_step_selection_non_nlp[
        "autoencoder_based_oversampling"
    ] = True
    titanic_auto_ml.blueprint_step_selection_non_nlp[
        "final_pca_dimensionality_reduction"
    ] = False
    titanic_auto_ml.blueprint_step_selection_non_nlp["scale_data"] = True

    titanic_auto_ml.ml_special_multiclass_full_processing_multimodel_max_voting()
    titanic_auto_ml.ml_special_multiclass_full_processing_multimodel_max_voting(val_df)
    # val_y_hat = titanic_auto_ml.predicted_classes["max_voting"]

    algorithms = [
        alg
        for alg, value in titanic_auto_ml.special_blueprint_algorithms.items()
        if value
    ]

    def get_matthews(algorithm):
        # Assess prediction quality on holdout data
        print(
            classification_report(
                pd.Series(val_df_target).astype(bool),
                titanic_auto_ml.predicted_classes[algorithm],
            )
        )
        try:
            matthews = matthews_corrcoef(
                pd.Series(val_df_target).astype(bool),
                titanic_auto_ml.predicted_classes[algorithm],
            )
        except Exception:
            print("Matthew failed.")
            matthews = 0
        print(matthews)

    for i in algorithms:
        print(f"---------Start evaluating {i}----------")
        get_matthews(i)
    finished = True
    assert finished is True


if __name__ == "__main__":
    test_ml_special_multiclass_full_processing_multimodel_max_voting("titanic")



import pandas as pd
from sklearn.metrics import mean_absolute_error

from e2eml.regression import regression_blueprints as rb

# track memory consumption in terminal: dmesg


def load_housingprices_data():
    """
    Load & preprocess Housing prices dataset. T
    :return: Several dataframes and series to be processed by blueprint.
    """
    data = pd.read_csv("housingprices_train.csv")

    def new_features(X):
        X["HasWoodDeck"] = (X["WoodDeckSF"] == 0) * 1
        X["HasOpenPorch"] = (X["OpenPorchSF"] == 0) * 1
        X["HasEnclosedPorch"] = (X["EnclosedPorch"] == 0) * 1
        X["Has3SsnPorch"] = (X["3SsnPorch"] == 0) * 1
        X["HasScreenPorch"] = (X["ScreenPorch"] == 0) * 1
        X["Total_Home_Quality"] = X["OverallQual"] + X["OverallCond"]
        X["TotalSF"] = X["TotalBsmtSF"] + X["1stFlrSF"] + X["2ndFlrSF"]
        X["TotalSquareFootage"] = (
            X["BsmtFinSF1"] + X["BsmtFinSF2"] + X["1stFlrSF"] + X["2ndFlrSF"]
        )
        X["HasPool"] = X["PoolArea"].apply(lambda x: 1 if x > 0 else 0)
        X["Has2ndFloor"] = X["2ndFlrSF"].apply(lambda x: 1 if x > 0 else 0)
        X["HasGarage"] = X["GarageArea"].apply(lambda x: 1 if x > 0 else 0)
        X["HasBsmt"] = X["TotalBsmtSF"].apply(lambda x: 1 if x > 0 else 0)
        X["HasFireplace"] = X["Fireplaces"].apply(lambda x: 1 if x > 0 else 0)
        return X

    data = new_features(data)
    print("Do dataframe splits.")
    test_df = data.head(1000).copy()
    val_df = data.tail(460).copy()
    val_df_target = val_df["SalePrice"].copy()
    del val_df["SalePrice"]
    test_target = "SalePrice"
    test_categorical_cols = [
        "MSZoning",
        "Street",
        "Alley",
        "LotShape",
        "LotFrontage",
        "Street",
        "LandContour",
        "Utilities",
        "LotConfig",
        "LandSlope",
        "Neighborhood",
        "Condition1",
        "Condition2",
        "BldgType",
        "HouseStyle",
        "RoofStyle",
        "RoofMatl",
        "Exterior1st",
        "Exterior2nd",
        "ExterQual",
        "ExterCond",
        "Foundation",
        "BsmtQual",
        "BsmtCond",
        "BsmtExposure",
        "BsmtFinType1",
        "BsmtFinType2",
        "Heating",
        "HeatingQC",
        "Electrical",
        "KitchenQual",
        "Functional",
        "FireplaceQU",
        "GarageType",
        "GarageYrBlt",
        "GarageFinish",
        "GarageQual",
        "GarageCond",
        "PavedDrive",
        "PoolQC",
        "Fence",
        "MiscFeature",
        "SaleType",
        "SaleCondition",
    ]
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def test_ml_special_regression_multiclass_full_processing_multimodel_avg_blender():
    (
        test_df,
        test_target,
        val_df,
        val_df_target,
        test_categorical_cols,
    ) = load_housingprices_data()
    titanic_auto_ml = rb.RegressionBluePrint(
        datasource=test_df,
        target_variable=test_target,
        categorical_columns=test_categorical_cols,
        preferred_training_mode="auto",
        tune_mode="accurate",
        ml_task="regression",
    )
    titanic_auto_ml.hyperparameter_tuning_rounds = {
        "xgboost": 10,
        "lgbm": 500,
        "tabnet": 3,
        "ngboost": 10,
        "sklearn_ensemble": 3,
        "catboost": 10,
        "ridge": 3,
        "bruteforce_random": 10,
        "elasticnet": 10,
        "autoencoder_based_oversampling": 20,
        "final_kernel_pca_dimensionality_reduction": 100,
        "final_pca_dimensionality_reduction": 20,
        "synthetic_data_augmentation": 100,
    }

    titanic_auto_ml.special_blueprint_algorithms = {
        "ridge": False,
        "xgboost": False,
        "ngboost": False,
        "lgbm": True,
        "tabnet": False,
        "vowpal_wabbit": False,
        "sklearn_ensemble": False,
        "catboost": False,
        "elasticnet": False,
    }
    titanic_auto_ml.blueprint_step_selection_non_nlp[
        "final_pca_dimensionality_reduction"
    ] = False
    titanic_auto_ml.blueprint_step_selection_non_nlp[
        "autoencoder_based_oversampling"
    ] = False
    titanic_auto_ml.blueprint_step_selection_non_nlp["scaling"] = False
    titanic_auto_ml.hyperparameter_tuning_sample_size = 800

    titanic_auto_ml.ml_special_regression_full_processing_multimodel_avg_blender()
    titanic_auto_ml.ml_special_regression_full_processing_multimodel_avg_blender(val_df)
    val_y_hat = titanic_auto_ml.predicted_values["blended_preds"]
    mae = mean_absolute_error(val_df_target, val_y_hat)
    finished = True
    assert finished is True
    assert mae >= 0


def test_ml_bp10_train_test_regression_full_processing_linear_reg():
    (
        test_df,
        test_target,
        val_df,
        val_df_target,
        test_categorical_cols,
    ) = load_housingprices_data()
    titanic_auto_ml = rb.RegressionBluePrint(
        datasource=test_df,
        target_variable=test_target,
        categorical_columns=test_categorical_cols,
        preferred_training_mode="auto",
        tune_mode="accurate",
    )
    titanic_auto_ml.ml_bp10_train_test_regression_full_processing_linear_reg()
    titanic_auto_ml.ml_bp10_train_test_regression_full_processing_linear_reg(val_df)
    val_y_hat = titanic_auto_ml.predicted_values["linear_regression"]
    mae = mean_absolute_error(val_df_target, val_y_hat)
    finished = True
    assert finished is True
    assert mae >= 0


if __name__ == "__main__":
    test_ml_special_regression_multiclass_full_processing_multimodel_avg_blender()
